from __future__ import annotations

import typing as t
from collections import defaultdict

from sqlglot import exp, generator, transforms
from sqlglot.dialects.dialect import (
    array_append_sql,
    array_concat_sql,
    date_delta_sql,
    datestrtodate_sql,
    groupconcat_sql,
    if_sql,
    inline_array_sql,
    map_date_part,
    max_or_greatest,
    min_or_least,
    no_make_interval_sql,
    no_timestamp_sql,
    rename_func,
    strposition_sql,
    timestampdiff_sql,
    timestamptrunc_sql,
    timestrtotime_sql,
    unit_to_str,
    var_map_sql,
)
from sqlglot.generator import unsupported_args
from sqlglot.helper import find_new_name, flatten, seq_get
from sqlglot.optimizer.scope import build_scope, find_all_in_scope
from sqlglot.parsers.snowflake import (
    RANKING_WINDOW_FUNCTIONS_WITH_FRAME,
    TIMESTAMP_TYPES,
    SnowflakeParser,
    build_object_construct,
)
from sqlglot.tokens import TokenType

if t.TYPE_CHECKING:
    from sqlglot._typing import E


def _build_datediff(args: list) -> exp.DateDiff:
    return exp.DateDiff(
        this=seq_get(args, 2),
        expression=seq_get(args, 1),
        unit=map_date_part(seq_get(args, 0)),
        date_part_boundary=True,
    )


def _build_date_time_add(expr_type: type[E]) -> t.Callable[[list], E]:
    def _builder(args: list) -> E:
        return expr_type(
            this=seq_get(args, 2),
            expression=seq_get(args, 1),
            unit=map_date_part(seq_get(args, 0)),
        )

    return _builder


def _regexpilike_sql(self: SnowflakeGenerator, expression: exp.RegexpILike) -> str:
    flag = expression.text("flag")

    if "i" not in flag:
        flag += "i"

    return self.func(
        "REGEXP_LIKE", expression.this, expression.expression, exp.Literal.string(flag)
    )


def _unqualify_pivot_columns(expression: exp.Expr) -> exp.Expr:
    """
    Snowflake doesn't allow columns referenced in UNPIVOT to be qualified,
    so we need to unqualify them. Same goes for ANY ORDER BY <column>.

    Example:
        >>> from sqlglot import parse_one
        >>> expr = parse_one("SELECT * FROM m_sales UNPIVOT(sales FOR month IN (m_sales.jan, feb, mar, april))")
        >>> print(_unqualify_pivot_columns(expr).sql(dialect="snowflake"))
        SELECT * FROM m_sales UNPIVOT(sales FOR month IN (jan, feb, mar, april))
    """
    if isinstance(expression, exp.Pivot):
        if expression.unpivot:
            expression = transforms.unqualify_columns(expression)
        else:
            for field in expression.fields:
                field_expr = seq_get(field.expressions if field else [], 0)

                if isinstance(field_expr, exp.PivotAny):
                    unqualified_field_expr = transforms.unqualify_columns(field_expr)
                    t.cast(exp.Expr, field).set("expressions", unqualified_field_expr, 0)

    return expression


def _flatten_structured_types_unless_iceberg(expression: exp.Expr) -> exp.Expr:
    assert isinstance(expression, exp.Create)

    def _flatten_structured_type(expression: exp.Expr) -> exp.Expr:
        if isinstance(expression, exp.DataType) and expression.this in exp.DataType.NESTED_TYPES:
            expression.set("expressions", None)
        return expression

    props = expression.args.get("properties")
    if isinstance(expression.this, exp.Schema) and not (props and props.find(exp.IcebergProperty)):
        for schema_expression in expression.this.expressions:
            if isinstance(schema_expression, exp.ColumnDef):
                column_type = schema_expression.kind
                if isinstance(column_type, exp.DataType):
                    column_type.transform(_flatten_structured_type, copy=False)

    return expression


def _unnest_generate_date_array(unnest: exp.Unnest) -> None:
    generate_date_array = unnest.expressions[0]
    start = generate_date_array.args.get("start")
    end = generate_date_array.args.get("end")
    step = generate_date_array.args.get("step")

    if not start or not end or not isinstance(step, exp.Interval) or step.name != "1":
        return

    unit = step.args.get("unit")

    unnest_alias = unnest.args.get("alias")
    if unnest_alias:
        unnest_alias = unnest_alias.copy()
        sequence_value_name = seq_get(unnest_alias.columns, 0) or "value"
    else:
        sequence_value_name = "value"

    # We'll add the next sequence value to the starting date and project the result
    date_add = _build_date_time_add(exp.DateAdd)(
        [unit, exp.cast(sequence_value_name, "int"), exp.cast(start, "date")]
    )

    # We use DATEDIFF to compute the number of sequence values needed
    number_sequence = SnowflakeParser.FUNCTIONS["ARRAY_GENERATE_RANGE"](
        [exp.Literal.number(0), _build_datediff([unit, start, end]) + 1]
    )

    unnest.set("expressions", [number_sequence])

    unnest_parent = unnest.parent
    if isinstance(unnest_parent, exp.Join):
        select = unnest_parent.parent
        if isinstance(select, exp.Select):
            replace_column_name = (
                sequence_value_name
                if isinstance(sequence_value_name, str)
                else sequence_value_name.name
            )

            scope = build_scope(select)
            if scope:
                for column in scope.columns:
                    if column.name.lower() == replace_column_name.lower():
                        column.replace(
                            date_add.as_(replace_column_name)
                            if isinstance(column.parent, exp.Select)
                            else date_add
                        )

            lateral = exp.Lateral(this=unnest_parent.this.pop())
            unnest_parent.replace(exp.Join(this=lateral))
    else:
        unnest.replace(
            exp.select(date_add.as_(sequence_value_name))
            .from_(unnest.copy())
            .subquery(unnest_alias)
        )


def _transform_generate_date_array(expression: exp.Expr) -> exp.Expr:
    if isinstance(expression, exp.Select):
        for generate_date_array in expression.find_all(exp.GenerateDateArray):
            parent = generate_date_array.parent

            # If GENERATE_DATE_ARRAY is used directly as an array (e.g passed into ARRAY_LENGTH), the transformed Snowflake
            # query is the following (it'll be unnested properly on the next iteration due to copy):
            # SELECT ref(GENERATE_DATE_ARRAY(...)) -> SELECT ref((SELECT ARRAY_AGG(*) FROM UNNEST(GENERATE_DATE_ARRAY(...))))
            if not isinstance(parent, exp.Unnest):
                unnest = exp.Unnest(expressions=[generate_date_array.copy()])
                generate_date_array.replace(
                    exp.select(exp.ArrayAgg(this=exp.Star())).from_(unnest).subquery()
                )

            if (
                isinstance(parent, exp.Unnest)
                and isinstance(parent.parent, (exp.From, exp.Join))
                and len(parent.expressions) == 1
            ):
                _unnest_generate_date_array(parent)

    return expression


def _regexpextract_sql(
    self: SnowflakeGenerator, expression: exp.RegexpExtract | exp.RegexpExtractAll
) -> str:
    # Other dialects don't support all of the following parameters, so we need to
    # generate default values as necessary to ensure the transpilation is correct
    group = expression.args.get("group")

    # To avoid generating all these default values, we set group to None if
    # it's 0 (also default value) which doesn't trigger the following chain
    if group and group.name == "0":
        group = None

    parameters = expression.args.get("parameters") or (group and exp.Literal.string("c"))
    occurrence = expression.args.get("occurrence") or (parameters and exp.Literal.number(1))
    position = expression.args.get("position") or (occurrence and exp.Literal.number(1))

    return self.func(
        "REGEXP_SUBSTR" if isinstance(expression, exp.RegexpExtract) else "REGEXP_SUBSTR_ALL",
        expression.this,
        expression.expression,
        position,
        occurrence,
        parameters,
        group,
    )


def _json_extract_value_array_sql(
    self: SnowflakeGenerator, expression: exp.JSONValueArray | exp.JSONExtractArray
) -> str:
    json_extract = exp.JSONExtract(this=expression.this, expression=expression.expression)
    ident = exp.to_identifier("x")

    if isinstance(expression, exp.JSONValueArray):
        this: exp.Expr = exp.cast(ident, to=exp.DType.VARCHAR)
    else:
        this = exp.ParseJSON(this=f"TO_JSON({ident})")

    transform_lambda = exp.Lambda(expressions=[ident], this=this)

    return self.func("TRANSFORM", json_extract, transform_lambda)


def _qualify_unnested_columns(expression: exp.Expr) -> exp.Expr:
    if isinstance(expression, exp.Select):
        scope = build_scope(expression)
        if not scope:
            return expression

        unnests = list(scope.find_all(exp.Unnest))

        if not unnests:
            return expression

        taken_source_names = set(scope.sources)
        column_source: dict[str, exp.Identifier] = {}
        unnest_to_identifier: dict[exp.Unnest, exp.Identifier] = {}

        unnest_identifier: exp.Identifier | None = None
        orig_expression = expression.copy()

        for unnest in unnests:
            if not isinstance(unnest.parent, (exp.From, exp.Join)):
                continue

            # Try to infer column names produced by an unnest operator. This is only possible
            # when we can peek into the (statically known) contents of the unnested value.
            unnest_columns: set[str] = set()
            for unnest_expr in unnest.expressions:
                if not isinstance(unnest_expr, exp.Array):
                    continue

                for array_expr in unnest_expr.expressions:
                    if not (
                        isinstance(array_expr, exp.Struct)
                        and array_expr.expressions
                        and all(
                            isinstance(struct_expr, exp.PropertyEQ)
                            for struct_expr in array_expr.expressions
                        )
                    ):
                        continue

                    unnest_columns.update(
                        struct_expr.this.name.lower() for struct_expr in array_expr.expressions
                    )
                    break

                if unnest_columns:
                    break

            unnest_alias = unnest.args.get("alias")
            if not unnest_alias:
                alias_name = find_new_name(taken_source_names, "value")
                taken_source_names.add(alias_name)

                # Produce a `TableAlias` AST similar to what is produced for BigQuery. This
                # will be corrected later, when we generate SQL for the `Unnest` AST node.
                aliased_unnest = exp.alias_(unnest, None, table=[alias_name])
                scope.replace(unnest, aliased_unnest)

                unnest_identifier = aliased_unnest.args["alias"].columns[0]
            else:
                alias_columns = getattr(unnest_alias, "columns", [])
                unnest_identifier = unnest_alias.this or seq_get(alias_columns, 0)

            if not isinstance(unnest_identifier, exp.Identifier):
                return orig_expression

            unnest_to_identifier[unnest] = unnest_identifier
            column_source.update({c.lower(): unnest_identifier for c in unnest_columns})

        for column in scope.columns:
            if column.table:
                continue

            table = column_source.get(column.name.lower())
            if (
                unnest_identifier
                and not table
                and len(scope.sources) == 1
                and column.name.lower() != unnest_identifier.name.lower()
            ):
                unnest_ancestor = column.find_ancestor(exp.Unnest, exp.Select)
                if isinstance(unnest_ancestor, exp.Unnest):
                    ancestor_identifier = unnest_to_identifier.get(unnest_ancestor)
                    if (
                        ancestor_identifier
                        and ancestor_identifier.name.lower() == unnest_identifier.name.lower()
                    ):
                        continue

                table = unnest_identifier

            column.set("table", table and table.copy())

    return expression


def _eliminate_dot_variant_lookup(expression: exp.Expr) -> exp.Expr:
    if isinstance(expression, exp.Select):
        # This transformation is used to facilitate transpilation of BigQuery `UNNEST` operations
        # to Snowflake. It should not affect roundtrip because `Unnest` nodes cannot be produced
        # by Snowflake's parser.
        #
        # Additionally, at the time of writing this, BigQuery is the only dialect that produces a
        # `TableAlias` node that only fills `columns` and not `this`, due to `UNNEST_COLUMN_ONLY`.
        unnest_aliases = set()
        for unnest in find_all_in_scope(expression, exp.Unnest):
            unnest_alias = unnest.args.get("alias")
            if (
                isinstance(unnest_alias, exp.TableAlias)
                and not unnest_alias.this
                and len(unnest_alias.columns) == 1
            ):
                unnest_aliases.add(unnest_alias.columns[0].name)

        if unnest_aliases:
            for c in find_all_in_scope(expression, exp.Column):
                if c.table in unnest_aliases:
                    bracket_lhs = c.args["table"]
                    bracket_rhs = exp.Literal.string(c.name)
                    bracket = exp.Bracket(this=bracket_lhs, expressions=[bracket_rhs])

                    if c.parent is expression:
                        # Retain column projection names by using aliases
                        c.replace(exp.alias_(bracket, c.this.copy()))
                    else:
                        c.replace(bracket)

    return expression


class SnowflakeGenerator(generator.Generator):
    SELECT_KINDS: tuple[str, ...] = ()
    PARAMETER_TOKEN = "$"
    MATCHED_BY_SOURCE = False
    SINGLE_STRING_INTERVAL = True
    JOIN_HINTS = False
    TABLE_HINTS = False
    QUERY_HINTS = False
    SUPPORTS_TABLE_COPY = False
    COLLATE_IS_FUNC = True
    LIMIT_ONLY_LITERALS = True
    JSON_KEY_VALUE_PAIR_SEP = ","
    INSERT_OVERWRITE = " OVERWRITE INTO"
    STRUCT_DELIMITER = ("(", ")")
    COPY_PARAMS_ARE_WRAPPED = False
    COPY_PARAMS_EQ_REQUIRED = True
    STAR_EXCEPT = "EXCLUDE"
    SUPPORTS_EXPLODING_PROJECTIONS = False
    ARRAY_CONCAT_IS_VAR_LEN = False
    SUPPORTS_CONVERT_TIMEZONE = True
    EXCEPT_INTERSECT_SUPPORT_ALL_CLAUSE = False
    SUPPORTS_MEDIAN = True
    ARRAY_SIZE_NAME = "ARRAY_SIZE"
    SUPPORTS_DECODE_CASE = True

    AFTER_HAVING_MODIFIER_TRANSFORMS = generator.AFTER_HAVING_MODIFIER_TRANSFORMS

    IS_BOOL_ALLOWED = False
    DIRECTED_JOINS = True
    SUPPORTS_UESCAPE = False
    TRY_SUPPORTED = False

    TRANSFORMS = {
        **generator.Generator.TRANSFORMS,
        exp.ApproxDistinct: rename_func("APPROX_COUNT_DISTINCT"),
        exp.ArgMax: rename_func("MAX_BY"),
        exp.ArgMin: rename_func("MIN_BY"),
        exp.Array: transforms.preprocess([transforms.inherit_struct_field_names]),
        exp.ArrayConcat: array_concat_sql("ARRAY_CAT"),
        exp.ArrayAppend: array_append_sql("ARRAY_APPEND"),
        exp.ArrayPrepend: array_append_sql("ARRAY_PREPEND"),
        exp.ArrayContains: lambda self, e: self.func(
            "ARRAY_CONTAINS",
            e.expression
            if e.args.get("ensure_variant") is False
            else exp.cast(e.expression, exp.DType.VARIANT, copy=False),
            e.this,
        ),
        exp.ArrayPosition: lambda self, e: self.func(
            "ARRAY_POSITION",
            e.expression,
            e.this,
        ),
        exp.ArrayIntersect: rename_func("ARRAY_INTERSECTION"),
        exp.ArrayOverlaps: rename_func("ARRAYS_OVERLAP"),
        exp.AtTimeZone: lambda self, e: self.func("CONVERT_TIMEZONE", e.args.get("zone"), e.this),
        exp.BitwiseOr: rename_func("BITOR"),
        exp.BitwiseXor: rename_func("BITXOR"),
        exp.BitwiseAnd: rename_func("BITAND"),
        exp.BitwiseAndAgg: rename_func("BITANDAGG"),
        exp.BitwiseOrAgg: rename_func("BITORAGG"),
        exp.BitwiseXorAgg: rename_func("BITXORAGG"),
        exp.BitwiseNot: rename_func("BITNOT"),
        exp.BitwiseLeftShift: rename_func("BITSHIFTLEFT"),
        exp.BitwiseRightShift: rename_func("BITSHIFTRIGHT"),
        exp.Create: transforms.preprocess([_flatten_structured_types_unless_iceberg]),
        exp.CurrentTimestamp: lambda self, e: (
            self.func("SYSDATE") if e.args.get("sysdate") else self.function_fallback_sql(e)
        ),
        exp.CurrentSchemas: lambda self, e: self.func("CURRENT_SCHEMAS"),
        exp.Localtime: lambda self, e: (
            self.func("CURRENT_TIME", e.this) if e.this else "CURRENT_TIME"
        ),
        exp.Localtimestamp: lambda self, e: (
            self.func("CURRENT_TIMESTAMP", e.this) if e.this else "CURRENT_TIMESTAMP"
        ),
        exp.DateAdd: date_delta_sql("DATEADD"),
        exp.DateDiff: date_delta_sql("DATEDIFF"),
        exp.DatetimeAdd: date_delta_sql("TIMESTAMPADD"),
        exp.DatetimeDiff: timestampdiff_sql,
        exp.DateStrToDate: datestrtodate_sql,
        exp.Decrypt: lambda self, e: self.func(
            f"{'TRY_' if e.args.get('safe') else ''}DECRYPT",
            e.this,
            e.args.get("passphrase"),
            e.args.get("aad"),
            e.args.get("encryption_method"),
        ),
        exp.DecryptRaw: lambda self, e: self.func(
            f"{'TRY_' if e.args.get('safe') else ''}DECRYPT_RAW",
            e.this,
            e.args.get("key"),
            e.args.get("iv"),
            e.args.get("aad"),
            e.args.get("encryption_method"),
            e.args.get("aead"),
        ),
        exp.DayOfMonth: rename_func("DAYOFMONTH"),
        exp.DayOfWeek: rename_func("DAYOFWEEK"),
        exp.DayOfWeekIso: rename_func("DAYOFWEEKISO"),
        exp.DayOfYear: rename_func("DAYOFYEAR"),
        exp.DotProduct: rename_func("VECTOR_INNER_PRODUCT"),
        exp.Explode: rename_func("FLATTEN"),
        exp.Extract: lambda self, e: self.func(
            "DATE_PART", map_date_part(e.this, self.dialect), e.expression
        ),
        exp.CosineDistance: rename_func("VECTOR_COSINE_SIMILARITY"),
        exp.EuclideanDistance: rename_func("VECTOR_L2_DISTANCE"),
        exp.HandlerProperty: lambda self, e: f"HANDLER = {self.sql(e, 'this')}",
        exp.FileFormatProperty: lambda self, e: (
            f"FILE_FORMAT=({self.expressions(e, 'expressions', sep=' ')})"
        ),
        exp.FromTimeZone: lambda self, e: self.func(
            "CONVERT_TIMEZONE", e.args.get("zone"), "'UTC'", e.this
        ),
        exp.GenerateSeries: lambda self, e: self.func(
            "ARRAY_GENERATE_RANGE",
            e.args["start"],
            e.args["end"] if e.args.get("is_end_exclusive") else e.args["end"] + 1,
            e.args.get("step"),
        ),
        exp.GetExtract: rename_func("GET"),
        exp.GroupConcat: lambda self, e: groupconcat_sql(self, e, sep=""),
        exp.If: if_sql(name="IFF", false_value="NULL"),
        exp.JSONArray: lambda self, e: self.func(
            "TO_VARIANT", self.func("ARRAY_CONSTRUCT", *e.expressions)
        ),
        exp.JSONExtractArray: _json_extract_value_array_sql,
        exp.JSONExtractScalar: lambda self, e: self.func(
            "JSON_EXTRACT_PATH_TEXT", e.this, e.expression
        ),
        exp.JSONKeys: rename_func("OBJECT_KEYS"),
        exp.JSONObject: lambda self, e: self.func("OBJECT_CONSTRUCT_KEEP_NULL", *e.expressions),
        exp.JSONPathRoot: lambda *_: "",
        exp.JSONValueArray: _json_extract_value_array_sql,
        exp.Levenshtein: unsupported_args("ins_cost", "del_cost", "sub_cost")(
            rename_func("EDITDISTANCE")
        ),
        exp.LocationProperty: lambda self, e: f"LOCATION={self.sql(e, 'this')}",
        exp.LogicalAnd: rename_func("BOOLAND_AGG"),
        exp.LogicalOr: rename_func("BOOLOR_AGG"),
        exp.Map: lambda self, e: var_map_sql(self, e, "OBJECT_CONSTRUCT"),
        exp.ManhattanDistance: rename_func("VECTOR_L1_DISTANCE"),
        exp.MakeInterval: no_make_interval_sql,
        exp.Max: max_or_greatest,
        exp.Min: min_or_least,
        exp.ParseJSON: lambda self, e: self.func(
            f"{'TRY_' if e.args.get('safe') else ''}PARSE_JSON", e.this
        ),
        exp.ToBinary: lambda self, e: self.func(
            f"{'TRY_' if e.args.get('safe') else ''}TO_BINARY", e.this, e.args.get("format")
        ),
        exp.ToBoolean: lambda self, e: self.func(
            f"{'TRY_' if e.args.get('safe') else ''}TO_BOOLEAN", e.this
        ),
        exp.ToDouble: lambda self, e: self.func(
            f"{'TRY_' if e.args.get('safe') else ''}TO_DOUBLE", e.this, e.args.get("format")
        ),
        exp.ToFile: lambda self, e: self.func(
            f"{'TRY_' if e.args.get('safe') else ''}TO_FILE", e.this, e.args.get("path")
        ),
        exp.JSONFormat: rename_func("TO_JSON"),
        exp.PartitionedByProperty: lambda self, e: f"PARTITION BY {self.sql(e, 'this')}",
        exp.PercentileCont: transforms.preprocess([transforms.add_within_group_for_percentiles]),
        exp.PercentileDisc: transforms.preprocess([transforms.add_within_group_for_percentiles]),
        exp.Pivot: transforms.preprocess([_unqualify_pivot_columns]),
        exp.RegexpExtract: _regexpextract_sql,
        exp.RegexpExtractAll: _regexpextract_sql,
        exp.RegexpILike: _regexpilike_sql,
        exp.RowAccessProperty: lambda self, e: self.rowaccessproperty_sql(e),
        exp.Select: transforms.preprocess(
            [
                transforms.eliminate_window_clause,
                transforms.eliminate_distinct_on,
                transforms.explode_projection_to_unnest(),
                transforms.eliminate_semi_and_anti_joins,
                _transform_generate_date_array,
                _qualify_unnested_columns,
                _eliminate_dot_variant_lookup,
            ]
        ),
        exp.SHA: rename_func("SHA1"),
        exp.SHA1Digest: rename_func("SHA1_BINARY"),
        exp.MD5Digest: rename_func("MD5_BINARY"),
        exp.MD5NumberLower64: rename_func("MD5_NUMBER_LOWER64"),
        exp.MD5NumberUpper64: rename_func("MD5_NUMBER_UPPER64"),
        exp.Hex: rename_func("HEX_ENCODE"),
        exp.LowerHex: rename_func("TO_CHAR"),
        exp.Skewness: rename_func("SKEW"),
        exp.StarMap: rename_func("OBJECT_CONSTRUCT"),
        exp.StartsWith: rename_func("STARTSWITH"),
        exp.EndsWith: rename_func("ENDSWITH"),
        exp.Rand: lambda self, e: self.func("RANDOM", e.this),
        exp.StrPosition: lambda self, e: strposition_sql(
            self, e, func_name="CHARINDEX", supports_position=True
        ),
        exp.StrToDate: lambda self, e: self.func("DATE", e.this, self.format_time(e)),
        exp.StringToArray: rename_func("STRTOK_TO_ARRAY"),
        exp.StrtokToArray: rename_func("STRTOK_TO_ARRAY"),
        exp.Stuff: rename_func("INSERT"),
        exp.StPoint: rename_func("ST_MAKEPOINT"),
        exp.TimeAdd: date_delta_sql("TIMEADD"),
        exp.TimeSlice: lambda self, e: self.func(
            "TIME_SLICE",
            e.this,
            e.expression,
            unit_to_str(e),
            e.args.get("kind"),
        ),
        exp.Timestamp: no_timestamp_sql,
        exp.TimestampAdd: date_delta_sql("TIMESTAMPADD"),
        exp.TimestampDiff: lambda self, e: self.func("TIMESTAMPDIFF", e.unit, e.expression, e.this),
        exp.TimestampTrunc: timestamptrunc_sql(),
        exp.TimeStrToTime: timestrtotime_sql,
        exp.TimeToUnix: lambda self, e: f"EXTRACT(epoch_second FROM {self.sql(e, 'this')})",
        exp.ToArray: rename_func("TO_ARRAY"),
        exp.ToChar: lambda self, e: self.function_fallback_sql(e),
        exp.TsOrDsAdd: date_delta_sql("DATEADD", cast=True),
        exp.TsOrDsDiff: date_delta_sql("DATEDIFF"),
        exp.TsOrDsToDate: lambda self, e: self.func(
            f"{'TRY_' if e.args.get('safe') else ''}TO_DATE", e.this, self.format_time(e)
        ),
        exp.TsOrDsToTime: lambda self, e: self.func(
            f"{'TRY_' if e.args.get('safe') else ''}TO_TIME", e.this, self.format_time(e)
        ),
        exp.Unhex: rename_func("HEX_DECODE_BINARY"),
        exp.UnixToTime: lambda self, e: self.func("TO_TIMESTAMP", e.this, e.args.get("scale")),
        exp.Uuid: rename_func("UUID_STRING"),
        exp.VarMap: lambda self, e: var_map_sql(self, e, "OBJECT_CONSTRUCT"),
        exp.Booland: rename_func("BOOLAND"),
        exp.Boolor: rename_func("BOOLOR"),
        exp.WeekOfYear: rename_func("WEEKISO"),
        exp.YearOfWeek: rename_func("YEAROFWEEK"),
        exp.YearOfWeekIso: rename_func("YEAROFWEEKISO"),
        exp.Xor: rename_func("BOOLXOR"),
        exp.ByteLength: rename_func("OCTET_LENGTH"),
        exp.Flatten: rename_func("ARRAY_FLATTEN"),
        exp.ArrayConcatAgg: lambda self, e: self.func("ARRAY_FLATTEN", exp.ArrayAgg(this=e.this)),
        exp.SHA2Digest: lambda self, e: self.func(
            "SHA2_BINARY", e.this, e.args.get("length") or exp.Literal.number(256)
        ),
    }

    def dynamicidentifier_sql(self, expression: exp.DynamicIdentifier) -> str:
        this = self.func("IDENTIFIER", expression.this)
        if "expressions" in expression.args:
            # `IDENTIFIER(...)` invoked as a function, e.g. `IDENTIFIER('my_func')(1, 2)`
            return self.func(this, *expression.expressions, normalize=False)
        return this

    def sortarray_sql(self, expression: exp.SortArray) -> str:
        asc = expression.args.get("asc")
        nulls_first = expression.args.get("nulls_first")
        if asc == exp.false() and nulls_first == exp.true():
            nulls_first = None
        return self.func("ARRAY_SORT", expression.this, asc, nulls_first)

    def nthvalue_sql(self, expression: exp.NthValue) -> str:
        result = self.func("NTH_VALUE", expression.this, expression.args.get("offset"))

        from_first = expression.args.get("from_first")

        if from_first is not None:
            if from_first:
                result = result + " FROM FIRST"
            else:
                result = result + " FROM LAST"

        return result

    SUPPORTED_JSON_PATH_PARTS = {
        exp.JSONPathKey,
        exp.JSONPathRoot,
        exp.JSONPathSubscript,
    }

    TYPE_MAPPING = {
        **generator.Generator.TYPE_MAPPING,
        exp.DType.BIGDECIMAL: "DOUBLE",
        exp.DType.JSON: "VARIANT",
        exp.DType.NESTED: "OBJECT",
        exp.DType.STRUCT: "OBJECT",
        exp.DType.TEXT: "VARCHAR",
    }

    TOKEN_MAPPING = {
        TokenType.AUTO_INCREMENT: "AUTOINCREMENT",
    }

    PROPERTIES_LOCATION = {
        **generator.Generator.PROPERTIES_LOCATION,
        exp.CredentialsProperty: exp.Properties.Location.POST_WITH,
        exp.LocationProperty: exp.Properties.Location.POST_WITH,
        exp.PartitionedByProperty: exp.Properties.Location.POST_SCHEMA,
        exp.RowAccessProperty: exp.Properties.Location.POST_SCHEMA,
        exp.SetProperty: exp.Properties.Location.UNSUPPORTED,
        exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    }

    UNSUPPORTED_VALUES_EXPRESSIONS: t.ClassVar = {
        exp.Map,
        exp.StarMap,
        exp.Struct,
        exp.VarMap,
    }

    RESPECT_IGNORE_NULLS_UNSUPPORTED_EXPRESSIONS = (exp.ArrayAgg,)

    def with_properties(self, properties: exp.Properties) -> str:
        return self.properties(properties, wrapped=False, prefix=self.sep(""), sep=" ")

    def values_sql(self, expression: exp.Values, values_as_table: bool = True) -> str:
        if expression.find(*self.UNSUPPORTED_VALUES_EXPRESSIONS):
            values_as_table = False

        return super().values_sql(expression, values_as_table=values_as_table)

    def datatype_sql(self, expression: exp.DataType) -> str:
        # Check if this is a FLOAT type nested inside a VECTOR type
        # VECTOR only accepts FLOAT (not DOUBLE), INT, and STRING as element types
        # https://docs.snowflake.com/en/sql-reference/data-types-vector
        if expression.is_type(exp.DType.DOUBLE):
            parent = expression.parent
            if isinstance(parent, exp.DataType) and parent.is_type(exp.DType.VECTOR):
                # Preserve FLOAT for VECTOR types instead of mapping to synonym DOUBLE
                return "FLOAT"

        expressions = expression.expressions
        if expressions and expression.is_type(*exp.DataType.STRUCT_TYPES):
            for field_type in expressions:
                # The correct syntax is OBJECT [ (<key> <value_type [NOT NULL] [, ...]) ]
                if isinstance(field_type, exp.DataType):
                    return "OBJECT"
                if (
                    isinstance(field_type, exp.ColumnDef)
                    and field_type.this
                    and field_type.this.is_string
                ):
                    # Doing OBJECT('foo' VARCHAR) is invalid snowflake Syntax. Moreover, besides
                    # converting 'foo' into an identifier, we also need to quote it because these
                    # keys are case-sensitive. For example:
                    #
                    # WITH t AS (SELECT OBJECT_CONSTRUCT('x', 'y') AS c) SELECT c:x FROM t -- correct
                    # WITH t AS (SELECT OBJECT_CONSTRUCT('x', 'y') AS c) SELECT c:X FROM t -- incorrect, returns NULL
                    field_type.this.replace(exp.to_identifier(field_type.name, quoted=True))

        return super().datatype_sql(expression)

    def tonumber_sql(self, expression: exp.ToNumber) -> str:
        precision = expression.args.get("precision")
        scale = expression.args.get("scale")

        default_precision = isinstance(precision, exp.Literal) and precision.name == "38"
        default_scale = isinstance(scale, exp.Literal) and scale.name == "0"

        if default_precision and default_scale:
            precision = None
            scale = None
        elif default_scale:
            scale = None

        func_name = "TRY_TO_NUMBER" if expression.args.get("safe") else "TO_NUMBER"

        return self.func(
            func_name,
            expression.this,
            expression.args.get("format"),
            precision,
            scale,
        )

    def timestampfromparts_sql(self, expression: exp.TimestampFromParts) -> str:
        milli = expression.args.get("milli")
        if milli is not None:
            milli_to_nano = milli.pop() * exp.Literal.number(1000000)
            expression.set("nano", milli_to_nano)

        return rename_func("TIMESTAMP_FROM_PARTS")(self, expression)

    def cast_sql(self, expression: exp.Cast, safe_prefix: str | None = None) -> str:
        if expression.is_type(exp.DType.GEOGRAPHY):
            return self.func("TO_GEOGRAPHY", expression.this)
        if expression.is_type(exp.DType.GEOMETRY):
            return self.func("TO_GEOMETRY", expression.this)

        return super().cast_sql(expression, safe_prefix=safe_prefix)

    def trycast_sql(self, expression: exp.TryCast) -> str:
        value = expression.this

        if value.type is None:
            from sqlglot.optimizer.annotate_types import annotate_types

            value = annotate_types(value, dialect=self.dialect)

        # Snowflake requires that TRY_CAST's value be a string
        # If TRY_CAST is being roundtripped (since Snowflake is the only dialect that sets "requires_string") or
        # if we can deduce that the value is a string, then we can generate TRY_CAST
        if expression.args.get("requires_string") or value.is_type(*exp.DataType.TEXT_TYPES):
            return super().trycast_sql(expression)

        return self.cast_sql(expression)

    def log_sql(self, expression: exp.Log) -> str:
        if not expression.expression:
            return self.func("LN", expression.this)

        return super().log_sql(expression)

    def greatest_sql(self, expression: exp.Greatest) -> str:
        name = "GREATEST_IGNORE_NULLS" if expression.args.get("ignore_nulls") else "GREATEST"
        return self.func(name, expression.this, *expression.expressions)

    def least_sql(self, expression: exp.Least) -> str:
        name = "LEAST_IGNORE_NULLS" if expression.args.get("ignore_nulls") else "LEAST"
        return self.func(name, expression.this, *expression.expressions)

    def generator_sql(self, expression: exp.Generator) -> str:
        args = []
        rowcount = expression.args.get("rowcount")
        timelimit = expression.args.get("timelimit")

        if rowcount:
            args.append(exp.Kwarg(this=exp.var("ROWCOUNT"), expression=rowcount))
        if timelimit:
            args.append(exp.Kwarg(this=exp.var("TIMELIMIT"), expression=timelimit))

        return self.func("GENERATOR", *args)

    def unnest_sql(self, expression: exp.Unnest) -> str:
        unnest_alias = expression.args.get("alias")
        offset = expression.args.get("offset")

        unnest_alias_columns = unnest_alias.columns if unnest_alias else []
        value = seq_get(unnest_alias_columns, 0) or exp.to_identifier("value")

        columns = [
            exp.to_identifier("seq"),
            exp.to_identifier("key"),
            exp.to_identifier("path"),
            offset.pop() if isinstance(offset, exp.Expr) else exp.to_identifier("index"),
            value,
            exp.to_identifier("this"),
        ]

        if unnest_alias:
            unnest_alias.set("columns", columns)
        else:
            unnest_alias = exp.TableAlias(this="_u", columns=columns)

        table_input = self.sql(expression.expressions[0])
        if not table_input.startswith("INPUT =>"):
            table_input = f"INPUT => {table_input}"

        expression_parent = expression.parent

        explode = (
            f"FLATTEN({table_input})"
            if isinstance(expression_parent, exp.Lateral)
            else f"TABLE(FLATTEN({table_input}))"
        )
        alias = self.sql(unnest_alias)
        alias = f" AS {alias}" if alias else ""
        value = (
            ""
            if isinstance(expression_parent, (exp.From, exp.Join, exp.Lateral))
            else f"{value} FROM "
        )

        return f"{value}{explode}{alias}"

    def undrop_sql(self, expression: exp.Undrop) -> str:
        this = self.sql(expression, "this")
        kind = expression.kind
        rename = self.sql(expression, "rename")
        rename = f" RENAME TO {rename}" if rename else ""
        return f"UNDROP {kind} {this}{rename}"

    def show_sql(self, expression: exp.Show) -> str:
        terse = "TERSE " if expression.args.get("terse") else ""
        iceberg = "ICEBERG " if expression.args.get("iceberg") else ""
        history = " HISTORY" if expression.args.get("history") else ""
        like = self.sql(expression, "like")
        like = f" LIKE {like}" if like else ""

        scope = self.sql(expression, "scope")
        scope = f" {scope}" if scope else ""

        scope_kind = self.sql(expression, "scope_kind")
        if scope_kind:
            scope_kind = f" IN {scope_kind}"

        starts_with = self.sql(expression, "starts_with")
        if starts_with:
            starts_with = f" STARTS WITH {starts_with}"

        limit = self.sql(expression, "limit")

        from_ = self.sql(expression, "from_")
        if from_:
            from_ = f" FROM {from_}"

        privileges = self.expressions(expression, key="privileges", flat=True)
        privileges = f" WITH PRIVILEGES {privileges}" if privileges else ""

        return f"SHOW {terse}{iceberg}{expression.name}{history}{like}{scope_kind}{scope}{starts_with}{limit}{from_}{privileges}"

    def rowaccessproperty_sql(self, expression: exp.RowAccessProperty) -> str:
        if not expression.this:
            return "ROW ACCESS"
        on = f" ON ({self.expressions(expression, flat=True)})" if expression.expressions else ""
        return f"WITH ROW ACCESS POLICY {self.sql(expression, 'this')}{on}"

    def describe_sql(self, expression: exp.Describe) -> str:
        kind_value = expression.args.get("kind") or "TABLE"

        properties = expression.args.get("properties")
        if properties:
            qualifier = self.expressions(properties, sep=" ")
            kind = f" {qualifier} {kind_value}"
        else:
            kind = f" {kind_value}"

        this = f" {self.sql(expression, 'this')}"
        expressions = self.expressions(expression, flat=True)
        expressions = f" {expressions}" if expressions else ""
        return f"DESCRIBE{kind}{this}{expressions}"

    def generatedasidentitycolumnconstraint_sql(
        self, expression: exp.GeneratedAsIdentityColumnConstraint
    ) -> str:
        start = expression.args.get("start")
        start = f" START {start}" if start else ""
        increment = expression.args.get("increment")
        increment = f" INCREMENT {increment}" if increment else ""

        order = expression.args.get("order")
        if order is not None:
            order_clause = " ORDER" if order else " NOORDER"
        else:
            order_clause = ""

        return f"AUTOINCREMENT{start}{increment}{order_clause}"

    def struct_sql(self, expression: exp.Struct) -> str:
        if len(expression.expressions) == 1:
            arg = expression.expressions[0]
            if arg.is_star or (isinstance(arg, exp.ILike) and arg.left.is_star):
                # Wildcard syntax: https://docs.snowflake.com/en/sql-reference/data-types-semistructured#object
                return f"{{{self.sql(expression.expressions[0])}}}"

        keys = []
        values = []

        for i, e in enumerate(expression.expressions):
            if isinstance(e, exp.PropertyEQ):
                keys.append(
                    exp.Literal.string(e.name) if isinstance(e.this, exp.Identifier) else e.this
                )
                values.append(e.expression)
            else:
                keys.append(exp.Literal.string(f"_{i}"))
                values.append(e)

        return self.func("OBJECT_CONSTRUCT", *flatten(zip(keys, values)))

    @unsupported_args("weight", "accuracy")
    def approxquantile_sql(self, expression: exp.ApproxQuantile) -> str:
        return self.func("APPROX_PERCENTILE", expression.this, expression.args.get("quantile"))

    def alterset_sql(self, expression: exp.AlterSet) -> str:
        exprs = self.expressions(expression, flat=True)
        exprs = f" {exprs}" if exprs else ""
        file_format = self.expressions(expression, key="file_format", flat=True, sep=" ")
        file_format = f" STAGE_FILE_FORMAT = ({file_format})" if file_format else ""
        copy_options = self.expressions(expression, key="copy_options", flat=True, sep=" ")
        copy_options = f" STAGE_COPY_OPTIONS = ({copy_options})" if copy_options else ""
        tag = self.expressions(expression, key="tag", flat=True)
        tag = f" TAG {tag}" if tag else ""

        return f"SET{exprs}{file_format}{copy_options}{tag}"

    def strtotime_sql(self, expression: exp.StrToTime):
        # target_type is stored as a DataType instance
        target_type = expression.args.get("target_type")

        # Get the type enum from DataType instance or from type annotation
        if isinstance(target_type, exp.DataType):
            type_enum = target_type.this
        elif expression.type:
            type_enum = expression.type.this
        else:
            type_enum = exp.DType.TIMESTAMP

        func_name = TIMESTAMP_TYPES.get(type_enum, "TO_TIMESTAMP")

        return self.func(
            f"{'TRY_' if expression.args.get('safe') else ''}{func_name}",
            expression.this,
            self.format_time(expression),
        )

    def timestampsub_sql(self, expression: exp.TimestampSub):
        return self.sql(
            exp.TimestampAdd(
                this=expression.this,
                expression=expression.expression * -1,
                unit=expression.unit,
            )
        )

    def jsonextract_sql(self, expression: exp.JSONExtract):
        this = expression.this

        # JSON strings are valid coming from other dialects such as BQ so
        # for these cases we PARSE_JSON preemptively
        if not isinstance(this, (exp.ParseJSON, exp.JSONExtract)) and not expression.args.get(
            "requires_json"
        ):
            this = exp.ParseJSON(this=this)

        return self.func(
            "GET_PATH",
            this,
            expression.expression,
        )

    def timetostr_sql(self, expression: exp.TimeToStr) -> str:
        this = expression.this
        if this.is_string:
            this = exp.cast(this, exp.DType.TIMESTAMP)

        return self.func("TO_CHAR", this, self.format_time(expression))

    def datesub_sql(self, expression: exp.DateSub) -> str:
        value = expression.expression
        if value:
            value.replace(value * (-1))
        else:
            self.unsupported("DateSub cannot be transpiled if the subtracted count is unknown")

        return date_delta_sql("DATEADD")(self, expression)

    def select_sql(self, expression: exp.Select) -> str:
        limit = expression.args.get("limit")
        offset = expression.args.get("offset")
        if offset and not limit:
            expression.limit(exp.Null(), copy=False)
        return super().select_sql(expression)

    def createable_sql(self, expression: exp.Create, locations: defaultdict) -> str:
        is_materialized = expression.find(exp.MaterializedProperty)
        copy_grants_property = expression.find(exp.CopyGrantsProperty)

        if expression.kind == "VIEW" and is_materialized and copy_grants_property:
            # For materialized views, COPY GRANTS is located *before* the columns list
            # This is in contrast to normal views where COPY GRANTS is located *after* the columns list
            # We default CopyGrantsProperty to POST_SCHEMA which means we need to output it POST_NAME if a materialized view is detected
            # ref: https://docs.snowflake.com/en/sql-reference/sql/create-materialized-view#syntax
            # ref: https://docs.snowflake.com/en/sql-reference/sql/create-view#syntax
            post_schema_properties = locations[exp.Properties.Location.POST_SCHEMA]
            post_schema_properties.pop(post_schema_properties.index(copy_grants_property))

            this_name = self.sql(expression.this, "this")
            copy_grants = self.sql(copy_grants_property)
            this_schema = self.schema_columns_sql(expression.this)
            this_schema = f"{self.sep()}{this_schema}" if this_schema else ""

            return f"{this_name}{self.sep()}{copy_grants}{this_schema}"

        return super().createable_sql(expression, locations)

    def arrayagg_sql(self, expression: exp.ArrayAgg) -> str:
        this = expression.this

        # If an ORDER BY clause is present, we need to remove it from ARRAY_AGG
        # and add it later as part of the WITHIN GROUP clause
        order = this if isinstance(this, exp.Order) else None
        if order:
            expression.set("this", order.this.pop())

        expr_sql = super().arrayagg_sql(expression)

        if order:
            expr_sql = self.sql(exp.WithinGroup(this=expr_sql, expression=order))

        return expr_sql

    def arraydistinct_sql(self, expression: exp.ArrayDistinct) -> str:
        if expression.args.get("check_null"):
            return self.func("ARRAY_DISTINCT", expression.this)
        return self.func("ARRAY_DISTINCT", exp.ArrayCompact(this=expression.this))

    def arraytostring_sql(self, expression: exp.ArrayToString) -> str:
        return self.func("ARRAY_TO_STRING", expression.this, expression.expression)

    def array_sql(self, expression: exp.Array) -> str:
        expressions = expression.expressions

        first_expr = seq_get(expressions, 0)
        if isinstance(first_expr, exp.Select):
            # SELECT AS STRUCT foo AS alias_foo -> ARRAY_AGG(OBJECT_CONSTRUCT('alias_foo', foo))
            if first_expr.text("kind").upper() == "STRUCT":
                object_construct_args = []
                for expr in first_expr.expressions:
                    # Alias case: SELECT AS STRUCT foo AS alias_foo -> OBJECT_CONSTRUCT('alias_foo', foo)
                    # Column case: SELECT AS STRUCT foo -> OBJECT_CONSTRUCT('foo', foo)
                    name = expr.this if isinstance(expr, exp.Alias) else expr

                    object_construct_args.extend([exp.Literal.string(expr.alias_or_name), name])

                array_agg = exp.ArrayAgg(this=build_object_construct(args=object_construct_args))

                first_expr.set("kind", None)
                first_expr.set("expressions", [array_agg])

                return self.sql(first_expr.subquery())

        return inline_array_sql(self, expression)

    def currentdate_sql(self, expression: exp.CurrentDate) -> str:
        zone = self.sql(expression, "this")
        if not zone:
            return super().currentdate_sql(expression)

        expr = exp.Cast(
            this=exp.ConvertTimezone(target_tz=zone, timestamp=exp.CurrentTimestamp()),
            to=exp.DataType(this=exp.DType.DATE),
        )
        return self.sql(expr)

    def dot_sql(self, expression: exp.Dot) -> str:
        this = expression.this

        if not this.type:
            from sqlglot.optimizer.annotate_types import annotate_types

            this = annotate_types(this, dialect=self.dialect)

        if not isinstance(this, exp.Dot) and this.is_type(exp.DType.STRUCT):
            # Generate colon notation for the top level STRUCT
            return f"{self.sql(this)}:{self.sql(expression, 'expression')}"

        return super().dot_sql(expression)

    def modelattribute_sql(self, expression: exp.ModelAttribute) -> str:
        return f"{self.sql(expression, 'this')}!{self.sql(expression, 'expression')}"

    def format_sql(self, expression: exp.Format) -> str:
        if expression.name.lower() == "%s" and len(expression.expressions) == 1:
            return self.func("TO_CHAR", expression.expressions[0])

        return self.function_fallback_sql(expression)

    def splitpart_sql(self, expression: exp.SplitPart) -> str:
        # Set part_index to 1 if missing
        if not expression.args.get("delimiter"):
            expression.set("delimiter", exp.Literal.string(" "))

        if not expression.args.get("part_index"):
            expression.set("part_index", exp.Literal.number(1))

        return rename_func("SPLIT_PART")(self, expression)

    def uniform_sql(self, expression: exp.Uniform) -> str:
        gen = expression.args.get("gen")
        seed = expression.args.get("seed")

        # From Databricks UNIFORM(min, max, seed) -> Wrap gen in RANDOM(seed)
        if seed:
            gen = exp.Rand(this=seed)

        # No gen argument (from Databricks 2-arg UNIFORM(min, max)) -> Add RANDOM()
        if not gen:
            gen = exp.Rand()

        return self.func("UNIFORM", expression.this, expression.expression, gen)

    def window_sql(self, expression: exp.Window) -> str:
        spec = expression.args.get("spec")
        this = expression.this

        if (
            (
                isinstance(this, RANKING_WINDOW_FUNCTIONS_WITH_FRAME)
                or (
                    isinstance(this, (exp.RespectNulls, exp.IgnoreNulls))
                    and isinstance(this.this, RANKING_WINDOW_FUNCTIONS_WITH_FRAME)
                )
            )
            and spec
            and (
                spec.text("kind").upper() == "ROWS"
                and spec.text("start").upper() == "UNBOUNDED"
                and spec.text("start_side").upper() == "PRECEDING"
                and spec.text("end").upper() == "UNBOUNDED"
                and spec.text("end_side").upper() == "FOLLOWING"
            )
        ):
            # omit the default window from window ranking functions
            expression.set("spec", None)
        return super().window_sql(expression)

    def filter_sql(self, expression: exp.Filter) -> str:
        # Snowflake doesn't support FILTER (WHERE cond), so we rewrite it into an
        # equivalent conditional aggregation, i.e. wrap the input values in an IFF
        agg = expression.this
        agg_arg = agg.this
        cond = expression.expression.this

        if isinstance(agg, exp.WithinGroup):
            # Ordered-set aggregates take their input from the ORDER BY key, so the
            # condition has to wrap that instead of the aggregate's own argument
            if isinstance(agg_arg, (exp.Mode, *exp.PERCENTILES)):
                for ordered in agg.expression.expressions:
                    key = ordered.this
                    key.replace(exp.If(this=cond.copy(), true=key.copy()))

                return self.sql(agg)

            # Besides the percentile functions, these are the only functions Snowflake
            # accepts WITHIN GROUP for, so anything else can't be rewritten correctly
            if isinstance(agg_arg, (exp.ArrayAgg, exp.GroupConcat)):
                agg_arg = agg_arg.this
            else:
                self.unsupported("Unable to rewrite FILTER into the aggregate's arguments")
                return self.sql(agg)

        # `COUNT(*/t.*) FILTER (WHERE cond)` counts qualifying rows, but a star can't be an IFF
        # argument: `IFF(cond, *, NULL)` expands to multiple columns once the table has 2+ of
        # them, which Snowflake rejects. Use its native COUNT_IF instead.
        if isinstance(agg, exp.Count) and agg_arg.is_star:
            return self.func("COUNT_IF", cond)

        # `DISTINCT` and `ORDER BY` are part of the aggregate's own argument list, so the
        # condition has to wrap the values underneath them rather than the whole clause --
        # `IFF(cond, DISTINCT x, NULL)` is not a call any dialect accepts.
        if isinstance(agg_arg, exp.Order):
            agg_arg = agg_arg.this

        if isinstance(agg_arg, exp.Distinct):
            targets = agg_arg.expressions
        else:
            targets = [agg_arg]

        for target in targets:
            target.replace(exp.If(this=cond.copy(), true=target.copy()))

        return self.sql(agg)

    def withingroup_sql(self, expression: exp.WithinGroup) -> str:
        # Snowflake's MODE doesn't support the ordered-set syntax, i.e. it only
        # accepts the value to aggregate as an argument: MODE(<expr>)
        if isinstance(expression.this, exp.Mode) and not expression.this.this:
            order = expression.expression
            if isinstance(order, exp.Order) and len(order.expressions) == 1:
                return self.sql(exp.Mode(this=order.expressions[0].this))

        return super().withingroup_sql(expression)



from __future__ import annotations

from sqlglot import exp, tokens
from sqlglot.dialects.dialect import Dialect
from sqlglot.generators.postgres import PostgresGenerator
from sqlglot.parsers.postgres import PostgresParser
from sqlglot.tokens import TokenType
from sqlglot.typing.postgres import EXPRESSION_METADATA


class Postgres(Dialect):
    EXPRESSION_METADATA = EXPRESSION_METADATA.copy()
    INDEX_OFFSET = 1
    # Normalizing `x IS NOT NULL` to `NOT x IS NULL` is unsafe due to row values,
    # e.g. `ROW(1, NULL) IS NOT NULL` is false whereas `NOT ROW(1, NULL) IS NULL` is true
    NORMALIZE_NOT_NULL = False
    TYPED_DIVISION = True
    CONCAT_COALESCE = True
    CONCAT_WS_COALESCE = True
    NULL_ORDERING = "nulls_are_large"
    SUPPORTS_LIMIT_ALL = True
    TIME_FORMAT = "'YYYY-MM-DD HH24:MI:SS'"
    TABLESAMPLE_SIZE_IS_PERCENT = True
    TABLES_REFERENCEABLE_AS_COLUMNS = True

    DEFAULT_FUNCTIONS_COLUMN_NAMES = {
        exp.ExplodingGenerateSeries: "generate_series",
    }

    TIME_MAPPING = {
        "d": "%u",  # 1-based day of week
        "D": "%u",  # 1-based day of week
        "dd": "%d",  # day of month
        "DD": "%d",  # day of month
        "ddd": "%j",  # zero padded day of year
        "DDD": "%j",  # zero padded day of year
        "FMDD": "%-d",  # - is no leading zero for Python; same for FM in postgres
        "FMDDD": "%-j",  # day of year
        "FMHH12": "%-I",  # 9
        "FMHH24": "%-H",  # 9
        "FMMI": "%-M",  # Minute
        "FMMM": "%-m",  # 1
        "FMSS": "%-S",  # Second
        "HH12": "%I",  # 09
        "HH24": "%H",  # 09
        "mi": "%M",  # zero padded minute
        "MI": "%M",  # zero padded minute
        "mm": "%m",  # 01
        "MM": "%m",  # 01
        "OF": "%z",  # utc offset
        "ss": "%S",  # zero padded second
        "SS": "%S",  # zero padded second
        "TMDay": "%A",  # TM is locale dependent
        "TMDy": "%a",
        "TMMon": "%b",  # Sep
        "TMMonth": "%B",  # September
        "day": "%Aenlower",  # tuesday
        "dy": "%aenlower",  # tue
        "TZ": "%Z",  # uppercase timezone name
        "US": "%f",  # zero padded microsecond
        "ww": "%U",  # 1-based week of year
        "WW": "%U",  # 1-based week of year
        "yy": "%y",  # 15
        "YY": "%y",  # 15
        "yyy": "%Ythree",  # 015
        "YYY": "%Ythree",  # 015
        "yyyy": "%Y",  # 2015
        "YYYY": "%Y",  # 2015
    }

    class Tokenizer(tokens.Tokenizer):
        BIT_STRINGS = [("b'", "'"), ("B'", "'")]
        HEX_STRINGS = [("x'", "'"), ("X'", "'")]
        BYTE_STRINGS = [("e'", "'"), ("E'", "'")]
        UNICODE_STRINGS = [("U&'", "'"), ("u&'", "'")]
        BYTE_STRING_ESCAPES = ["'", "\\"]
        HEREDOC_STRINGS = ["$"]

        HEREDOC_TAG_IS_IDENTIFIER = True
        HEREDOC_STRING_ALTERNATIVE = TokenType.PARAMETER

        KEYWORDS = {
            **tokens.Tokenizer.KEYWORDS,
            "~": TokenType.RLIKE,
            "@@": TokenType.DAT,
            "@?": TokenType.AT_QMARK,
            "@>": TokenType.AT_GT,
            "<@": TokenType.LT_AT,
            "?&": TokenType.QMARK_AMP,
            "?|": TokenType.QMARK_PIPE,
            "#-": TokenType.HASH_DASH,
            "|/": TokenType.PIPE_SLASH,
            "||/": TokenType.DPIPE_SLASH,
            "BEGIN": TokenType.BEGIN,
            "BIGSERIAL": TokenType.BIGSERIAL,
            "CSTRING": TokenType.PSEUDO_TYPE,
            "DECLARE": TokenType.COMMAND,
            "DO": TokenType.COMMAND,
            "EXEC": TokenType.COMMAND,
            "HSTORE": TokenType.HSTORE,
            "INT8": TokenType.BIGINT,
            "MONEY": TokenType.MONEY,
            "NAME": TokenType.NAME,
            "OID": TokenType.OBJECT_IDENTIFIER,
            "ONLY": TokenType.ONLY,
            "POINT": TokenType.POINT,
            "REFRESH": TokenType.COMMAND,
            "REINDEX": TokenType.COMMAND,
            "RESET": TokenType.COMMAND,
            "SERIAL": TokenType.SERIAL,
            "SMALLSERIAL": TokenType.SMALLSERIAL,
            "TEMP": TokenType.TEMPORARY,
            "TYPE": TokenType.TYPE,
            "REGCLASS": TokenType.OBJECT_IDENTIFIER,
            "REGCOLLATION": TokenType.OBJECT_IDENTIFIER,
            "REGCONFIG": TokenType.OBJECT_IDENTIFIER,
            "REGDICTIONARY": TokenType.OBJECT_IDENTIFIER,
            "REGNAMESPACE": TokenType.OBJECT_IDENTIFIER,
            "REGOPER": TokenType.OBJECT_IDENTIFIER,
            "REGOPERATOR": TokenType.OBJECT_IDENTIFIER,
            "REGPROC": TokenType.OBJECT_IDENTIFIER,
            "REGPROCEDURE": TokenType.OBJECT_IDENTIFIER,
            "REGROLE": TokenType.OBJECT_IDENTIFIER,
            "REGTYPE": TokenType.OBJECT_IDENTIFIER,
            "FLOAT": TokenType.DOUBLE,
            "XML": TokenType.XML,
            "VARIADIC": TokenType.VARIADIC,
            "INOUT": TokenType.INOUT,
        }
        KEYWORDS.pop("/*+")
        KEYWORDS.pop("DIV")

        SINGLE_TOKENS = {
            **tokens.Tokenizer.SINGLE_TOKENS,
            "$": TokenType.HEREDOC_STRING,
        }

        VAR_SINGLE_TOKENS = {"$"}

    Parser = PostgresParser

    Generator = PostgresGenerator



from __future__ import annotations

from sqlglot import exp, transforms
from sqlglot.dialects.dialect import (
    merge_without_target_sql,
    trim_sql,
    timestrtotime_sql,
    groupconcat_sql,
    rename_func,
)
from sqlglot.generators.presto import PrestoGenerator, amend_exploded_column_table


class TrinoGenerator(PrestoGenerator):
    EXCEPT_INTERSECT_SUPPORT_ALL_CLAUSE = True
    DECLARE_DEFAULT_ASSIGNMENT = "DEFAULT"
    PROPERTIES_LOCATION = {
        **PrestoGenerator.PROPERTIES_LOCATION,
        exp.LocationProperty: exp.Properties.Location.POST_WITH,
    }

    TRANSFORMS = {
        **PrestoGenerator.TRANSFORMS,
        exp.ArraySum: lambda self, e: (
            f"REDUCE({self.sql(e, 'this')}, 0, (acc, x) -> acc + x, acc -> acc)"
        ),
        exp.ArrayUniqueAgg: lambda self, e: f"ARRAY_AGG(DISTINCT {self.sql(e, 'this')})",
        exp.CurrentVersion: rename_func("VERSION"),
        exp.FromISO8601TimestampNanos: rename_func("FROM_ISO8601_TIMESTAMP_NANOS"),
        exp.GroupConcat: lambda self, e: groupconcat_sql(self, e, on_overflow=True),
        exp.LocationProperty: lambda self, e: self.property_sql(e),
        exp.Merge: merge_without_target_sql,
        exp.Select: transforms.preprocess(
            [
                transforms.eliminate_qualify,
                transforms.eliminate_distinct_on,
                transforms.explode_projection_to_unnest(1),
                transforms.eliminate_semi_and_anti_joins,
                amend_exploded_column_table,
            ]
        ),
        exp.StabilityProperty: lambda self, e: (
            "DETERMINISTIC" if e.name == "IMMUTABLE" else "NOT DETERMINISTIC"
        ),
        exp.TimeStrToTime: lambda self, e: timestrtotime_sql(self, e, include_precision=True),
        exp.Trim: trim_sql,
    }

    SUPPORTED_JSON_PATH_PARTS = {
        exp.JSONPathKey,
        exp.JSONPathRoot,
        exp.JSONPathSubscript,
    }

    def functionspecification_sql(self, expression: exp.FunctionSpecification) -> str:
        characteristics = expression.args.get("characteristics")
        characteristics_sql = (
            self.properties(characteristics, prefix=" ", sep=" ", wrapped=False)
            if characteristics
            else ""
        )
        properties = expression.args.get("properties")
        with_sql = f" {self.with_properties(properties)}" if properties else ""
        body = self.sql(expression, "expression")
        return f"FUNCTION {self.sql(expression, 'this')}{characteristics_sql}{with_sql} {body}"

    def ifblock_sql(self, expression: exp.IfBlock) -> str:
        # ELSEIF chains are nested into `false` at parse time (see
        # TrinoParser._parse_routine_if), so this flattens them back out rather
        # than recursing on ifblock_sql itself, which would re-wrap each link in
        # its own IF ... END IF.
        branches: list[str] = []
        node: exp.Expr | None = expression

        while isinstance(node, exp.IfBlock):
            keyword = "IF" if not branches else "ELSEIF"
            branches.append(f"{keyword} {self.sql(node, 'this')} THEN {self.sql(node, 'true')};")
            node = node.args.get("false")

        if node is not None:
            branches.append(f"ELSE {self.sql(node)};")

        return f"{' '.join(branches)} END IF"

    def casestatement_sql(self, expression: exp.CaseStatement) -> str:
        # Mirrors case_sql, using `;`-terminated statement bodies and END CASE
        # instead of a single value expression per branch and bare END.
        this = self.sql(expression, "this")
        branches = [f"CASE {this}" if this else "CASE"]

        for node in expression.args["ifs"]:
            branches.append(f"WHEN {self.sql(node, 'this')} THEN {self.sql(node, 'true')};")

        default = expression.args.get("default")
        if default:
            branches.append(f"ELSE {self.sql(default)};")

        branches.append("END CASE")
        return " ".join(branches)

    def whileblock_sql(self, expression: exp.WhileBlock) -> str:
        label = expression.args.get("label")
        label_sql = f"{self.sql(label)}: " if label else ""
        condition = self.sql(expression, "this")
        body = self.sql(expression, "body")
        return f"{label_sql}WHILE {condition} DO {body}; END WHILE"

    def loopblock_sql(self, expression: exp.LoopBlock) -> str:
        label = expression.args.get("label")
        label_sql = f"{self.sql(label)}: " if label else ""
        body = self.sql(expression, "body")
        return f"{label_sql}LOOP {body}; END LOOP"

    def repeatblock_sql(self, expression: exp.RepeatBlock) -> str:
        label = expression.args.get("label")
        label_sql = f"{self.sql(label)}: " if label else ""
        body = self.sql(expression, "body")
        until = self.sql(expression, "until")
        return f"{label_sql}REPEAT {body}; UNTIL {until} END REPEAT"

    def leave_sql(self, expression: exp.Leave) -> str:
        return f"LEAVE {self.sql(expression, 'this')}"

    def iterate_sql(self, expression: exp.Iterate) -> str:
        return f"ITERATE {self.sql(expression, 'this')}"

    def jsonextract_sql(self, expression: exp.JSONExtract) -> str:
        if not expression.args.get("json_query"):
            return super().jsonextract_sql(expression)

        json_path = self.sql(expression, "expression")
        option = self.sql(expression, "option")
        option = f" {option}" if option else ""

        quote = self.sql(expression, "quote")
        quote = f" {quote}" if quote else ""

        on_condition = self.sql(expression, "on_condition")
        on_condition = f" {on_condition}" if on_condition else ""

        return self.func(
            "JSON_QUERY",
            expression.this,
            json_path + option + quote + on_condition,
        )



from __future__ import annotations

import typing as t

from sqlglot import expressions as exp
from types import ModuleType


StackVal = tuple[t.Any, t.Optional[int], t.Optional[str], bool]


INDEX = "i"
ARG_KEY = "k"
IS_ARRAY = "a"
CLASS = "c"
TYPE = "t"
COMMENTS = "o"
META = "m"
VALUE = "v"
DATA_TYPE = "DataType.Type"


def dump(expression: exp.Expr) -> list[dict[str, t.Any]]:
    """
    Dump an Expr into a JSON serializable List.
    """
    i = 0
    payloads: list[dict[str, t.Any]] = []
    stack: list[StackVal] = [(expression, None, None, False)]

    while stack:
        node, index, arg_key, is_array = stack.pop()

        payload: dict[str, t.Any] = {}

        if index is not None:
            payload[INDEX] = index
        if arg_key is not None:
            payload[ARG_KEY] = arg_key
        if is_array:
            payload[IS_ARRAY] = is_array

        payloads.append(payload)

        if hasattr(node, "parent"):
            klass = node.__class__.__qualname__

            if node.__class__.__module__ != exp.__name__:
                klass = f"{node.__module__}.{klass}"

            payload[CLASS] = klass

            if node.type and node.type is not node:
                payload[TYPE] = dump(node.type)
            if node.comments:
                payload[COMMENTS] = node.comments
            if node._meta is not None:
                payload[META] = node._meta
            if node.args:
                for k, vs in reversed(node.args.items()):
                    if type(vs) is list:
                        for v in reversed(vs):
                            stack.append((v, i, k, True))
                    elif vs is not None:
                        stack.append((vs, i, k, False))
        elif type(node) is exp.DType:
            payload[CLASS] = DATA_TYPE
            payload[VALUE] = node.value
        else:
            payload[VALUE] = node

        i += 1

    return payloads


def load(
    payloads: list[dict[str, t.Any]] | None,
) -> exp.Expr | exp.DType | None:
    """
    Load a list of dicts generated by dump into an Expr.
    """

    if not payloads:
        return None

    payload, *tail = payloads
    root = _load(payload)
    nodes: list[object] = [root]
    for payload in tail:
        if CLASS in payload:
            node: object = _load(payload)
        else:
            node = payload[VALUE]

        nodes.append(node)
        parent: exp.Expr = nodes[payload[INDEX]]
        arg_key: str = payload[ARG_KEY]

        if payload.get(IS_ARRAY):
            parent.append(arg_key, node)
        else:
            parent.set(arg_key, node)

    return root


def _load(payload: dict[str, t.Any]) -> exp.Expr | exp.DType:
    class_name: str = payload[CLASS]

    if class_name == DATA_TYPE:
        return exp.DType(payload[VALUE])
    module: ModuleType
    if "." in class_name:
        module_path, class_name = class_name.rsplit(".", maxsplit=1)
        module = __import__(module_path, fromlist=[class_name])
    else:
        module = exp

    expression = getattr(module, class_name)()
    expression._type = load(payload.get(TYPE))
    expression.comments = payload.get(COMMENTS)
    expression._meta = payload.get(META)
    return expression



from __future__ import annotations


from sqlglot import exp
from sqlglot.trie import new_trie
from sqlglot.dialects.dialect import (
    build_formatted_time,
    build_json_extract_path,
)
from sqlglot.dialects.mysql import MySQL
from sqlglot.helper import seq_get
from sqlglot.parsers.mysql import MySQLParser, _show_parser
from sqlglot.tokens import TokenType


def cast_to_time6(expression: exp.Expr | None, time_type: exp.DType = exp.DType.TIME) -> exp.Cast:
    return exp.Cast(
        this=expression,
        to=time_type.into_expr(expressions=[exp.DataTypeParam(this=exp.Literal.number(6))]),
    )


class SingleStoreParser(MySQLParser):
    FUNCTIONS = {
        **MySQLParser.FUNCTIONS,
        "TO_DATE": build_formatted_time(exp.TsOrDsToDate),
        "TO_TIMESTAMP": build_formatted_time(exp.StrToTime),
        "TO_CHAR": build_formatted_time(exp.ToChar),
        "STR_TO_DATE": build_formatted_time(exp.StrToDate, "mysql"),
        "DATE_FORMAT": build_formatted_time(exp.TimeToStr, "mysql"),
        # The first argument of following functions is converted to TIME(6)
        # This is needed because exp.TimeToStr is converted to DATE_FORMAT
        # which interprets the first argument as DATETIME and fails to parse
        # string literals like '12:05:47' without a date part.
        "TIME_FORMAT": lambda args: exp.TimeToStr(
            this=cast_to_time6(seq_get(args, 0)),
            format=MySQL.format_time(seq_get(args, 1)),
        ),
        "HOUR": lambda args: exp.cast(
            exp.TimeToStr(
                this=cast_to_time6(seq_get(args, 0)),
                format=MySQL.format_time(exp.Literal.string("%k")),
            ),
            exp.DType.INT,
        ),
        "MICROSECOND": lambda args: exp.cast(
            exp.TimeToStr(
                this=cast_to_time6(seq_get(args, 0)),
                format=MySQL.format_time(exp.Literal.string("%f")),
            ),
            exp.DType.INT,
        ),
        "SECOND": lambda args: exp.cast(
            exp.TimeToStr(
                this=cast_to_time6(seq_get(args, 0)),
                format=MySQL.format_time(exp.Literal.string("%s")),
            ),
            exp.DType.INT,
        ),
        "MINUTE": lambda args: exp.cast(
            exp.TimeToStr(
                this=cast_to_time6(seq_get(args, 0)),
                format=MySQL.format_time(exp.Literal.string("%i")),
            ),
            exp.DType.INT,
        ),
        "MONTHNAME": lambda args: exp.TimeToStr(
            this=seq_get(args, 0),
            format=MySQL.format_time(exp.Literal.string("%M")),
        ),
        "WEEKDAY": lambda args: exp.paren(exp.DayOfWeek(this=seq_get(args, 0)) + 5, copy=False) % 7,
        "UNIX_TIMESTAMP": exp.StrToUnix.from_arg_list,
        "FROM_UNIXTIME": build_formatted_time(exp.UnixToTime, "mysql"),
        "TIME_BUCKET": lambda args: exp.DateBin(
            this=seq_get(args, 0),
            expression=seq_get(args, 1),
            origin=seq_get(args, 2),
        ),
        "BSON_EXTRACT_BSON": build_json_extract_path(exp.JSONBExtract),
        "BSON_EXTRACT_STRING": build_json_extract_path(exp.JSONBExtractScalar, json_type="STRING"),
        "BSON_EXTRACT_DOUBLE": build_json_extract_path(exp.JSONBExtractScalar, json_type="DOUBLE"),
        "BSON_EXTRACT_BIGINT": build_json_extract_path(exp.JSONBExtractScalar, json_type="BIGINT"),
        "JSON_EXTRACT_JSON": build_json_extract_path(exp.JSONExtract),
        "JSON_EXTRACT_STRING": build_json_extract_path(exp.JSONExtractScalar, json_type="STRING"),
        "JSON_EXTRACT_DOUBLE": build_json_extract_path(exp.JSONExtractScalar, json_type="DOUBLE"),
        "JSON_EXTRACT_BIGINT": build_json_extract_path(exp.JSONExtractScalar, json_type="BIGINT"),
        "JSON_ARRAY_CONTAINS_STRING": lambda args: exp.JSONArrayContains(
            this=seq_get(args, 1),
            expression=seq_get(args, 0),
            json_type="STRING",
        ),
        "JSON_ARRAY_CONTAINS_DOUBLE": lambda args: exp.JSONArrayContains(
            this=seq_get(args, 1),
            expression=seq_get(args, 0),
            json_type="DOUBLE",
        ),
        "JSON_ARRAY_CONTAINS_JSON": lambda args: exp.JSONArrayContains(
            this=seq_get(args, 1),
            expression=seq_get(args, 0),
            json_type="JSON",
        ),
        "JSON_KEYS": lambda args: exp.JSONKeys(
            this=seq_get(args, 0),
            expressions=args[1:],
        ),
        "JSON_PRETTY": exp.JSONFormat.from_arg_list,
        "JSON_BUILD_ARRAY": lambda args: exp.JSONArray(expressions=args),
        "JSON_BUILD_OBJECT": lambda args: exp.JSONObject(expressions=args),
        "DATE": exp.Date.from_arg_list,
        "DAYNAME": lambda args: exp.TimeToStr(
            this=seq_get(args, 0),
            format=MySQL.format_time(exp.Literal.string("%W")),
        ),
        "TIMESTAMPDIFF": lambda args: exp.TimestampDiff(
            this=seq_get(args, 2),
            expression=seq_get(args, 1),
            unit=seq_get(args, 0),
        ),
        "APPROX_COUNT_DISTINCT": exp.Hll.from_arg_list,
        "APPROX_PERCENTILE": lambda args, dialect: exp.ApproxQuantile(
            this=seq_get(args, 0),
            quantile=seq_get(args, 1),
            error_tolerance=seq_get(args, 2),
        ),
        "VARIANCE": exp.VariancePop.from_arg_list,
        "INSTR": exp.Contains.from_arg_list,
        "REGEXP_MATCH": lambda args: exp.RegexpExtractAll(
            this=seq_get(args, 0),
            expression=seq_get(args, 1),
            parameters=seq_get(args, 2),
        ),
        "REGEXP_SUBSTR": lambda args: exp.RegexpExtract(
            this=seq_get(args, 0),
            expression=seq_get(args, 1),
            position=seq_get(args, 2),
            occurrence=seq_get(args, 3),
            parameters=seq_get(args, 4),
        ),
        "REDUCE": lambda args: exp.Reduce(
            initial=seq_get(args, 0),
            this=seq_get(args, 1),
            merge=seq_get(args, 2),
        ),
    }

    FUNCTION_PARSERS = {
        **MySQLParser.FUNCTION_PARSERS,
        "JSON_AGG": lambda self: exp.JSONArrayAgg(
            this=self._parse_term(),
            order=self._parse_order(),
        ),
    }

    NO_PAREN_FUNCTIONS = {
        **MySQLParser.NO_PAREN_FUNCTIONS,
        TokenType.UTC_DATE: exp.UtcDate,
        TokenType.UTC_TIME: exp.UtcTime,
        TokenType.UTC_TIMESTAMP: exp.UtcTimestamp,
    }

    CAST_COLUMN_OPERATORS = {TokenType.COLON_GT, TokenType.NCOLON_GT}

    COLUMN_OPERATORS = {
        **MySQLParser.COLUMN_OPERATORS,
        TokenType.COLON_GT: lambda self, this, to: self.expression(exp.Cast(this=this, to=to)),
        TokenType.NCOLON_GT: lambda self, this, to: self.expression(exp.TryCast(this=this, to=to)),
        TokenType.DCOLON: lambda self, this, path: build_json_extract_path(exp.JSONExtract)(
            [this, exp.Literal.string(path.name)]
        ),
        TokenType.DCOLONDOLLAR: lambda self, this, path: build_json_extract_path(
            exp.JSONExtractScalar, json_type="STRING"
        )([this, exp.Literal.string(path.name)]),
        TokenType.DCOLONPERCENT: lambda self, this, path: build_json_extract_path(
            exp.JSONExtractScalar, json_type="DOUBLE"
        )([this, exp.Literal.string(path.name)]),
        TokenType.DCOLONQMARK: lambda self, this, path: self.expression(
            exp.JSONExists(this=this, path=path.name, from_dcolonqmark=True)
        ),
    }
    COLUMN_OPERATORS = {
        k: v
        for k, v in COLUMN_OPERATORS.items()
        if k
        not in (
            TokenType.ARROW,
            TokenType.DARROW,
            TokenType.HASH_ARROW,
            TokenType.DHASH_ARROW,
            TokenType.PLACEHOLDER,
        )
    }

    SHOW_PARSERS = {
        **MySQLParser.SHOW_PARSERS,
        "AGGREGATES": _show_parser("AGGREGATES"),
        "CDC EXTRACTOR POOL": _show_parser("CDC EXTRACTOR POOL"),
        "CREATE AGGREGATE": _show_parser("CREATE AGGREGATE", target=True),
        "CREATE PIPELINE": _show_parser("CREATE PIPELINE", target=True),
        "CREATE PROJECTION": _show_parser("CREATE PROJECTION", target=True),
        "DATABASE STATUS": _show_parser("DATABASE STATUS"),
        "DISTRIBUTED_PLANCACHE STATUS": _show_parser("DISTRIBUTED_PLANCACHE STATUS"),
        "FULLTEXT SERVICE METRICS LOCAL": _show_parser("FULLTEXT SERVICE METRICS LOCAL"),
        "FULLTEXT SERVICE METRICS FOR NODE": _show_parser(
            "FULLTEXT SERVICE METRICS FOR NODE", target=True
        ),
        "FULLTEXT SERVICE STATUS": _show_parser("FULLTEXT SERVICE STATUS"),
        "FUNCTIONS": _show_parser("FUNCTIONS"),
        "GROUPS": _show_parser("GROUPS"),
        "GROUPS FOR ROLE": _show_parser("GROUPS FOR ROLE", target=True),
        "GROUPS FOR USER": _show_parser("GROUPS FOR USER", target=True),
        "INDEXES": _show_parser("INDEX", target="FROM"),
        "KEYS": _show_parser("INDEX", target="FROM"),
        "LINKS": _show_parser("LINKS", target="ON"),
        "LOAD ERRORS": _show_parser("LOAD ERRORS"),
        "LOAD WARNINGS": _show_parser("LOAD WARNINGS"),
        "PARTITIONS": _show_parser("PARTITIONS", target="ON"),
        "PIPELINES": _show_parser("PIPELINES"),
        "PLAN": _show_parser("PLAN", target=True),
        "PLANCACHE": _show_parser("PLANCACHE"),
        "PROCEDURES": _show_parser("PROCEDURES"),
        "PROJECTIONS": _show_parser("PROJECTIONS", target="ON TABLE"),
        "REPLICATION STATUS": _show_parser("REPLICATION STATUS"),
        "REPRODUCTION": _show_parser("REPRODUCTION"),
        "RESOURCE POOLS": _show_parser("RESOURCE POOLS"),
        "ROLES": _show_parser("ROLES"),
        "ROLES FOR USER": _show_parser("ROLES FOR USER", target=True),
        "ROLES FOR GROUP": _show_parser("ROLES FOR GROUP", target=True),
        "STATUS EXTENDED": _show_parser("STATUS EXTENDED"),
        "USERS": _show_parser("USERS"),
        "USERS FOR ROLE": _show_parser("USERS FOR ROLE", target=True),
        "USERS FOR GROUP": _show_parser("USERS FOR GROUP", target=True),
    }

    SHOW_TRIE = new_trie(key.split(" ") for key in SHOW_PARSERS)

    ALTER_PARSERS = {
        **MySQLParser.ALTER_PARSERS,
        "CHANGE": lambda self: self.expression(
            exp.RenameColumn(this=self._parse_column(), to=self._parse_column())
        ),
    }

    def _parse_vector_expressions(self, expressions: list[exp.Expr]) -> list[exp.Expr]:
        type_name = expressions[1].name.upper()
        if type_name in self.dialect.VECTOR_TYPE_ALIASES:
            type_name = self.dialect.VECTOR_TYPE_ALIASES[type_name]

        return [exp.DataType.build(type_name, dialect=self.dialect), expressions[0]]



from __future__ import annotations

from sqlglot import exp, transforms
from sqlglot.generators.tsql import TSQLGenerator
from sqlglot.optimizer.scope import find_in_scope


def _cap_data_type_precision(expression: exp.DataType, max_precision: int = 6) -> exp.DataType:
    """
    Cap the precision of to a maximum of `max_precision` digits.
    If no precision is specified, default to `max_precision`.
    """

    precision_param = expression.find(exp.DataTypeParam)

    if precision_param and precision_param.this.is_int:
        current_precision = precision_param.this.to_py()
        target_precision = min(current_precision, max_precision)
    else:
        target_precision = max_precision

    return exp.DataType(
        this=expression.this,
        expressions=[exp.DataTypeParam(this=exp.Literal.number(target_precision))],
    )


def _add_default_precision_to_varchar(expression: exp.Expr) -> exp.Expr:
    """Transform function to add VARCHAR(MAX) or CHAR(MAX) for cross-dialect conversion."""
    if (
        isinstance(expression, exp.Create)
        and expression.kind == "TABLE"
        and isinstance(expression.this, exp.Schema)
    ):
        for column in expression.this.expressions:
            if isinstance(column, exp.ColumnDef):
                column_type = column.kind
                if (
                    isinstance(column_type, exp.DataType)
                    and column_type.this in (exp.DType.VARCHAR, exp.DType.CHAR)
                    and not column_type.expressions
                ):
                    # For transpilation, VARCHAR/CHAR without precision becomes VARCHAR(MAX)/CHAR(MAX)
                    column_type.set("expressions", [exp.var("MAX")])

    return expression


class FabricGenerator(TSQLGenerator):
    # Fabric-specific type mappings - override T-SQL types that aren't supported
    # Reference: https://learn.microsoft.com/en-us/fabric/data-warehouse/data-types
    TYPE_MAPPING = {
        **TSQLGenerator.TYPE_MAPPING,
        exp.DType.DATETIME: "DATETIME2",
        exp.DType.DECIMAL: "DECIMAL",
        exp.DType.IMAGE: "VARBINARY",
        exp.DType.INT: "INT",
        exp.DType.JSON: "VARCHAR",
        exp.DType.MONEY: "DECIMAL",
        exp.DType.NCHAR: "CHAR",
        exp.DType.NVARCHAR: "VARCHAR",
        exp.DType.ROWVERSION: "ROWVERSION",
        exp.DType.SMALLDATETIME: "DATETIME2",
        exp.DType.SMALLMONEY: "DECIMAL",
        exp.DType.TIMESTAMP: "DATETIME2",
        exp.DType.TIMESTAMPNTZ: "DATETIME2",
        exp.DType.TIMESTAMPTZ: "DATETIME2",
        exp.DType.TINYINT: "SMALLINT",
        exp.DType.UTINYINT: "SMALLINT",
        exp.DType.UUID: "UNIQUEIDENTIFIER",
        exp.DType.XML: "VARCHAR",
    }

    TRANSFORMS = {
        **TSQLGenerator.TRANSFORMS,
        exp.Create: transforms.preprocess([_add_default_precision_to_varchar]),
    }

    def datatype_sql(self, expression: exp.DataType) -> str:
        # Check if this is a temporal type that needs precision handling. Fabric limits temporal
        # types to max 6 digits precision. When no precision is specified, we default to 6 digits.
        if expression.is_type(*exp.DataType.TEMPORAL_TYPES) and expression.this != exp.DType.DATE:
            # Create a new expression with the capped precision
            expression = _cap_data_type_precision(expression)

        return super().datatype_sql(expression)

    def cast_sql(self, expression: exp.Cast, safe_prefix: str | None = None) -> str:
        # Cast to DATETIMEOFFSET if inside an AT TIME ZONE expression
        # https://learn.microsoft.com/en-us/sql/t-sql/data-types/datetimeoffset-transact-sql#microsoft-fabric-support
        if expression.is_type(exp.DType.TIMESTAMPTZ):
            at_time_zone = expression.find_ancestor(exp.AtTimeZone, exp.Select)

            # Return normal cast, if the expression is not in an AT TIME ZONE context
            if not isinstance(at_time_zone, exp.AtTimeZone):
                return super().cast_sql(expression, safe_prefix)

            # Get the precision from the original TIMESTAMPTZ cast and cap it to 6
            capped_data_type = _cap_data_type_precision(expression.to, max_precision=6)
            precision = capped_data_type.find(exp.DataTypeParam)
            precision_value = precision.this.to_py() if precision and precision.this.is_int else 6

            # Do the cast explicitly to bypass sqlglot's default handling
            datetimeoffset = f"CAST({expression.this} AS DATETIMEOFFSET({precision_value}))"

            return self.sql(datetimeoffset)

        return super().cast_sql(expression, safe_prefix)

    def attimezone_sql(self, expression: exp.AtTimeZone) -> str:
        # Wrap the AT TIME ZONE expression in a cast to DATETIME2 if it contains a TIMESTAMPTZ
        ## https://learn.microsoft.com/en-us/sql/t-sql/data-types/datetimeoffset-transact-sql#microsoft-fabric-support
        timestamptz_cast = find_in_scope(expression, exp.Cast)
        if timestamptz_cast and timestamptz_cast.to.is_type(exp.DType.TIMESTAMPTZ):
            # Get the precision from the original TIMESTAMPTZ cast and cap it to 6
            data_type = timestamptz_cast.to
            capped_data_type = _cap_data_type_precision(data_type, max_precision=6)
            precision_param = capped_data_type.find(exp.DataTypeParam)
            precision = precision_param.this.to_py() if precision_param else 6

            # Generate the AT TIME ZONE expression (which will handle the inner cast conversion)
            at_time_zone_sql = super().attimezone_sql(expression)

            # Wrap it in an outer cast to DATETIME2
            return f"CAST({at_time_zone_sql} AS DATETIME2({precision}))"

        return super().attimezone_sql(expression)

    def unixtotime_sql(self, expression: exp.UnixToTime) -> str:
        scale = expression.args.get("scale")
        timestamp = expression.this

        if scale not in (None, exp.UnixToTime.SECONDS):
            self.unsupported(f"UnixToTime scale {scale} is not supported by Fabric")
            return ""

        # Convert unix timestamp (seconds) to microseconds and round to avoid decimals
        microseconds = timestamp * exp.Literal.number("1e6")
        rounded = exp.func("round", microseconds, 0)
        rounded_ms_as_bigint = exp.cast(rounded, exp.DType.BIGINT)

        # Create the base datetime as '1970-01-01' cast to DATETIME2(6)
        epoch_start = exp.cast("'1970-01-01'", "datetime2(6)", dialect="fabric")

        dateadd = exp.DateAdd(
            this=epoch_start,
            expression=rounded_ms_as_bigint,
            unit=exp.Literal.string("MICROSECONDS"),
        )
        return self.sql(dateadd)



from __future__ import annotations

import typing as t

from sqlglot import exp, parser
from sqlglot.helper import seq_get
from sqlglot.tokens import TokenType
from sqlglot.trie import new_trie


class TeradataParser(parser.Parser):
    TABLESAMPLE_CSV = True
    VALUES_FOLLOWED_BY_PAREN = False

    TABLE_ALIAS_TOKENS = parser.Parser.TABLE_ALIAS_TOKENS | {
        TokenType.ANTI,
        TokenType.SEMI,
    }

    CHARSET_TRANSLATORS: t.ClassVar = {
        "GRAPHIC_TO_KANJISJIS",
        "GRAPHIC_TO_LATIN",
        "GRAPHIC_TO_UNICODE",
        "GRAPHIC_TO_UNICODE_PadSpace",
        "KANJI1_KanjiEBCDIC_TO_UNICODE",
        "KANJI1_KanjiEUC_TO_UNICODE",
        "KANJI1_KANJISJIS_TO_UNICODE",
        "KANJI1_SBC_TO_UNICODE",
        "KANJISJIS_TO_GRAPHIC",
        "KANJISJIS_TO_LATIN",
        "KANJISJIS_TO_UNICODE",
        "LATIN_TO_GRAPHIC",
        "LATIN_TO_KANJISJIS",
        "LATIN_TO_UNICODE",
        "LOCALE_TO_UNICODE",
        "UNICODE_TO_GRAPHIC",
        "UNICODE_TO_GRAPHIC_PadGraphic",
        "UNICODE_TO_GRAPHIC_VarGraphic",
        "UNICODE_TO_KANJI1_KanjiEBCDIC",
        "UNICODE_TO_KANJI1_KanjiEUC",
        "UNICODE_TO_KANJI1_KANJISJIS",
        "UNICODE_TO_KANJI1_SBC",
        "UNICODE_TO_KANJISJIS",
        "UNICODE_TO_LATIN",
        "UNICODE_TO_LOCALE",
        "UNICODE_TO_UNICODE_FoldSpace",
        "UNICODE_TO_UNICODE_Fullwidth",
        "UNICODE_TO_UNICODE_Halfwidth",
        "UNICODE_TO_UNICODE_NFC",
        "UNICODE_TO_UNICODE_NFD",
        "UNICODE_TO_UNICODE_NFKC",
        "UNICODE_TO_UNICODE_NFKD",
    }

    FUNC_TOKENS = parser.Parser.FUNC_TOKENS - {TokenType.REPLACE}

    STATEMENT_PARSERS = {
        **parser.Parser.STATEMENT_PARSERS,
        TokenType.DATABASE: lambda self: self.expression(
            exp.Use(this=self._parse_table(schema=False))
        ),
        TokenType.REPLACE: lambda self: self._parse_create(),
        TokenType.LOCK: lambda self: self._parse_locking_statement(),
    }

    def _parse_locking_statement(self) -> exp.LockingStatement:
        # Reuse exp.LockingProperty parsing for the lock kind, type etc
        locking_property = self._parse_locking()
        wrapped_query = self._parse_select()

        if not wrapped_query:
            self.raise_error("Expected SELECT statement after LOCKING clause")

        return self.expression(
            exp.LockingStatement(this=locking_property, expression=wrapped_query)
        )

    SET_PARSERS = {
        **parser.Parser.SET_PARSERS,
        "QUERY_BAND": lambda self: self._parse_query_band(),
    }

    SET_TRIE = new_trie(key.split(" ") for key in SET_PARSERS)

    FUNCTION_PARSERS = {
        **parser.Parser.FUNCTION_PARSERS,
        # https://docs.teradata.com/r/SQL-Functions-Operators-Exprs-and-Predicates/June-2017/Data-Type-Conversions/TRYCAST
        "TRYCAST": parser.Parser.FUNCTION_PARSERS["TRY_CAST"],
        "RANGE_N": lambda self: self._parse_rangen(),
        "TRANSLATE": lambda self: self._parse_translate(),
    }

    FUNCTIONS = {
        **parser.Parser.FUNCTIONS,
        "CARDINALITY": exp.ArraySize.from_arg_list,
        "RANDOM": lambda args: exp.Rand(lower=seq_get(args, 0), upper=seq_get(args, 1)),
    }

    EXPONENT = {
        TokenType.DSTAR: exp.Pow,
    }

    def _parse_translate(self) -> exp.TranslateCharacters:
        this = self._parse_assignment()
        self._match(TokenType.USING)
        self._match_texts(self.CHARSET_TRANSLATORS)

        return self.expression(
            exp.TranslateCharacters(
                this=this,
                expression=self._prev.text.upper(),
                with_error=self._match_text_seq("WITH", "ERROR"),
            )
        )

    # FROM before SET in Teradata UPDATE syntax
    # https://docs.teradata.com/r/Enterprise_IntelliFlex_VMware/Teradata-VantageTM-SQL-Data-Manipulation-Language-17.20/Statement-Syntax/UPDATE/UPDATE-Syntax-Basic-Form-FROM-Clause
    def _parse_update(self) -> exp.Update:
        return self.expression(
            exp.Update(
                this=self._parse_table(alias_tokens=self.UPDATE_ALIAS_TOKENS),
                from_=self._parse_from(joins=True),
                expressions=self._match(TokenType.SET) and self._parse_csv(self._parse_equality),
                where=self._parse_where(),
            )
        )

    def _parse_rangen(self):
        this = self._parse_id_var()
        self._match(TokenType.BETWEEN)

        expressions = self._parse_csv(self._parse_assignment)
        each = self._match_text_seq("EACH") and self._parse_assignment()

        return self.expression(exp.RangeN(this=this, expressions=expressions, each=each))

    def _parse_query_band(self) -> exp.QueryBand:
        # Parse: SET QUERY_BAND = 'key=value;key2=value2;' FOR SESSION|TRANSACTION
        # Also supports: SET QUERY_BAND = 'key=value;' UPDATE FOR SESSION|TRANSACTION
        # Also supports: SET QUERY_BAND = NONE FOR SESSION|TRANSACTION
        self._match(TokenType.EQ)

        # Handle both string literals and NONE keyword
        if self._match_text_seq("NONE"):
            query_band_string: exp.Expr | None = exp.Var(this="NONE")
        else:
            query_band_string = self._parse_string()

        update = self._match_text_seq("UPDATE")
        self._match_text_seq("FOR")

        # Handle scope - can be SESSION, TRANSACTION, VOLATILE, or SESSION VOLATILE
        if self._match_text_seq("SESSION", "VOLATILE"):
            scope = "SESSION VOLATILE"
        elif self._match_texts(("SESSION", "TRANSACTION")):
            scope = self._prev.text.upper()
        else:
            scope = None

        return self.expression(exp.QueryBand(this=query_band_string, scope=scope, update=update))

    def _parse_index_params(self) -> exp.IndexParameters:
        this = super()._parse_index_params()

        if this.args.get("on"):
            this.set("on", None)
            self._retreat(self._index - 2)
        return this

    def _parse_function(
        self,
        functions: dict[str, t.Callable] | None = None,
        anonymous: bool = False,
        optional_parens: bool = True,
        any_token: bool = False,
    ) -> exp.Expr | None:
        # Teradata uses a `(FORMAT <format_string>)` clause after column references to
        # override the output format. When we see this pattern we do not
        # parse it as a function call.  The syntax is documented at
        # https://docs.teradata.com/r/Enterprise_IntelliFlex_VMware/SQL-Data-Types-and-Literals/Data-Type-Formats-and-Format-Phrases/FORMAT
        if (
            self._next
            and self._next.token_type == TokenType.L_PAREN
            and self._index + 2 < len(self._tokens)
            and self._tokens[self._index + 2].token_type == TokenType.FORMAT
        ):
            return None

        return super()._parse_function(
            functions=functions,
            anonymous=anonymous,
            optional_parens=optional_parens,
            any_token=any_token,
        )

    def _parse_column_ops(self, this: exp.Expr | None) -> exp.Expr | None:
        this = super()._parse_column_ops(this)

        if self._match_pair(TokenType.L_PAREN, TokenType.FORMAT):
            # `(FORMAT <format_string>)` after a column specifies a Teradata format override.
            # See https://docs.teradata.com/r/Enterprise_IntelliFlex_VMware/SQL-Data-Types-and-Literals/Data-Type-Formats-and-Format-Phrases/FORMAT
            fmt_string = self._parse_string()
            self._match_r_paren()

            this = self.expression(exp.FormatPhrase(this=this, format=fmt_string))

        return this



from __future__ import annotations

import datetime
import typing as t

from sqlglot import exp, generator
from sqlglot.dialects.dialect import (
    arg_max_or_min_no_count,
    inline_array_sql,
    jarowinkler_similarity,
    json_extract_segments,
    json_path_key_only_name,
    length_or_char_length_sql,
    no_pivot_sql,
    rename_func,
    remove_from_array_using_filter,
    sha256_sql,
    strposition_sql,
    var_map_sql,
    weekstart_unit_to_str,
    unit_to_var,
    trim_sql,
    sha2_digest_sql,
)
from sqlglot.generator import unsupported_args
from sqlglot.helper import is_int
from collections import defaultdict

DATETIME_DELTA = t.Union[exp.DateAdd, exp.DateDiff, exp.DateSub, exp.TimestampSub, exp.TimestampAdd]


def _unix_to_time_sql(self: ClickHouseGenerator, expression: exp.UnixToTime) -> str:
    scale = expression.args.get("scale")
    timestamp = expression.this

    if scale in (None, exp.UnixToTime.SECONDS):
        return self.func("fromUnixTimestamp", exp.cast(timestamp, exp.DType.BIGINT))
    if scale == exp.UnixToTime.MILLIS:
        return self.func("fromUnixTimestamp64Milli", exp.cast(timestamp, exp.DType.BIGINT))
    if scale == exp.UnixToTime.MICROS:
        return self.func("fromUnixTimestamp64Micro", exp.cast(timestamp, exp.DType.BIGINT))
    if scale == exp.UnixToTime.NANOS:
        return self.func("fromUnixTimestamp64Nano", exp.cast(timestamp, exp.DType.BIGINT))

    return self.func(
        "fromUnixTimestamp",
        exp.cast(exp.Div(this=timestamp, expression=exp.func("POW", 10, scale)), exp.DType.BIGINT),
    )


def _lower_func(sql: str) -> str:
    index = sql.index("(")
    return sql[:index].lower() + sql[index:]


def _quantile_sql(self: ClickHouseGenerator, expression: exp.Quantile) -> str:
    quantile = expression.args["quantile"]
    args = f"({self.sql(expression, 'this')})"

    if isinstance(quantile, exp.Array):
        func = self.func("quantiles", *quantile)
    else:
        func = self.func("quantile", quantile)

    return func + args


def _datetime_delta_sql(name: str) -> t.Callable[[generator.Generator, DATETIME_DELTA], str]:
    def _delta_sql(self: generator.Generator, expression: DATETIME_DELTA) -> str:
        if not expression.unit:
            return rename_func(name)(self, expression)

        return self.func(
            name,
            unit_to_var(expression),
            expression.expression,
            expression.this,
            expression.args.get("zone"),
        )

    return _delta_sql


def _timestrtotime_sql(self: ClickHouseGenerator, expression: exp.TimeStrToTime):
    ts = expression.this

    tz = expression.args.get("zone")
    if tz and isinstance(ts, exp.Literal):
        # Clickhouse will not accept timestamps that include a UTC offset, so we must remove them.
        # The first step to removing is parsing the string with `datetime.datetime.fromisoformat`.
        #
        # In python <3.11, `fromisoformat()` can only parse timestamps of millisecond (3 digit)
        # or microsecond (6 digit) precision. It will error if passed any other number of fractional
        # digits, so we extract the fractional seconds and pad to 6 digits before parsing.
        ts_string = ts.name.strip()

        # separate [date and time] from [fractional seconds and UTC offset]
        ts_parts = ts_string.split(".")
        if len(ts_parts) == 2:
            # separate fractional seconds and UTC offset
            offset_sep = "+" if "+" in ts_parts[1] else "-"
            ts_frac_parts = ts_parts[1].split(offset_sep)
            num_frac_parts = len(ts_frac_parts)

            # pad to 6 digits if fractional seconds present
            ts_frac_parts[0] = ts_frac_parts[0].ljust(6, "0")
            ts_string = "".join(
                [
                    ts_parts[0],  # date and time
                    ".",
                    ts_frac_parts[0],  # fractional seconds
                    offset_sep if num_frac_parts > 1 else "",
                    ts_frac_parts[1] if num_frac_parts > 1 else "",  # utc offset (if present)
                ]
            )

        # return literal with no timezone, eg turn '2020-01-01 12:13:14-08:00' into '2020-01-01 12:13:14'
        # this is because Clickhouse encodes the timezone as a data type parameter and throws an error if
        # it's part of the timestamp string
        ts_without_tz = (
            datetime.datetime.fromisoformat(ts_string).replace(tzinfo=None).isoformat(sep=" ")
        )
        ts = exp.Literal.string(ts_without_tz)

    # Non-nullable DateTime64 with microsecond precision
    expressions = [exp.DataTypeParam(this=tz)] if tz else []
    datatype = exp.DType.DATETIME64.into_expr(
        expressions=[exp.DataTypeParam(this=exp.Literal.number(6)), *expressions],
        nullable=False,
    )

    return self.sql(exp.cast(ts, datatype, dialect=self.dialect))


def _map_sql(self: ClickHouseGenerator, expression: exp.Map | exp.VarMap) -> str:
    if not (expression.parent and expression.parent.arg_key == "settings"):
        return _lower_func(var_map_sql(self, expression))

    keys = expression.args.get("keys")
    values = expression.args.get("values")

    if not isinstance(keys, exp.Array) or not isinstance(values, exp.Array):
        self.unsupported("Cannot convert array columns into map.")
        return ""

    args = []
    for key, value in zip(keys.expressions, values.expressions):
        args.append(f"{self.sql(key)}: {self.sql(value)}")

    csv_args = ", ".join(args)

    return f"{{{csv_args}}}"


def _json_cast_sql(self: ClickHouseGenerator, expression: exp.JSONCast) -> str:
    this = self.sql(expression, "this")
    to = expression.to
    to_sql = self.sql(to)

    if to.expressions:
        to_sql = self.sql(exp.to_identifier(to_sql))

    return f"{this}.:{to_sql}"


class ClickHouseGenerator(generator.Generator):
    SELECT_KINDS: tuple[str, ...] = ()
    TRY_SUPPORTED = False
    SUPPORTS_UESCAPE = False
    SUPPORTS_DECODE_CASE = False

    AFTER_HAVING_MODIFIER_TRANSFORMS = generator.AFTER_HAVING_MODIFIER_TRANSFORMS

    QUERY_HINTS = False
    STRUCT_DELIMITER = ("(", ")")
    NVL2_SUPPORTED = False
    ALTER_SET_TYPE = "TYPE"
    SUPPORTS_ALTER_COLUMN_IF_EXISTS = True
    TABLESAMPLE_REQUIRES_PARENS = False
    TABLESAMPLE_SIZE_IS_ROWS = False
    TABLESAMPLE_KEYWORDS = "SAMPLE"
    LAST_DAY_SUPPORTS_DATE_PART = False
    CAN_IMPLEMENT_ARRAY_ANY = True
    SUPPORTS_TO_NUMBER = False
    JOIN_HINTS = False
    TABLE_HINTS = False
    GROUPINGS_SEP = ""
    SET_OP_MODIFIERS = False
    ARRAY_SIZE_NAME = "LENGTH"
    WRAP_DERIVED_VALUES = False
    AUTO_REFRESH_BARE_INTERVALS = True

    STRING_TYPE_MAPPING: t.ClassVar = {
        exp.DType.BLOB: "String",
        exp.DType.CHAR: "String",
        exp.DType.LONGBLOB: "String",
        exp.DType.LONGTEXT: "String",
        exp.DType.MEDIUMBLOB: "String",
        exp.DType.MEDIUMTEXT: "String",
        exp.DType.TINYBLOB: "String",
        exp.DType.TINYTEXT: "String",
        exp.DType.TEXT: "String",
        exp.DType.VARBINARY: "String",
        exp.DType.VARCHAR: "String",
    }

    SUPPORTED_JSON_PATH_PARTS = {
        exp.JSONPathKey,
        exp.JSONPathRoot,
        exp.JSONPathSubscript,
    }

    TYPE_MAPPING = {
        **generator.Generator.TYPE_MAPPING,
        exp.DType.BLOB: "String",
        exp.DType.CHAR: "String",
        exp.DType.LONGBLOB: "String",
        exp.DType.LONGTEXT: "String",
        exp.DType.MEDIUMBLOB: "String",
        exp.DType.MEDIUMTEXT: "String",
        exp.DType.TINYBLOB: "String",
        exp.DType.TINYTEXT: "String",
        exp.DType.TEXT: "String",
        exp.DType.VARBINARY: "String",
        exp.DType.VARCHAR: "String",
        exp.DType.ARRAY: "Array",
        exp.DType.BOOLEAN: "Bool",
        exp.DType.BIGINT: "Int64",
        exp.DType.DATE32: "Date32",
        exp.DType.DATETIME: "DateTime",
        exp.DType.DATETIME2: "DateTime",
        exp.DType.SMALLDATETIME: "DateTime",
        exp.DType.DATETIME64: "DateTime64",
        exp.DType.DECIMAL: "Decimal",
        exp.DType.DECIMAL32: "Decimal32",
        exp.DType.DECIMAL64: "Decimal64",
        exp.DType.DECIMAL128: "Decimal128",
        exp.DType.DECIMAL256: "Decimal256",
        exp.DType.TIMESTAMP: "DateTime",
        exp.DType.TIMESTAMPNTZ: "DateTime",
        exp.DType.TIMESTAMPTZ: "DateTime",
        exp.DType.DOUBLE: "Float64",
        exp.DType.ENUM: "Enum",
        exp.DType.ENUM8: "Enum8",
        exp.DType.ENUM16: "Enum16",
        exp.DType.FIXEDSTRING: "FixedString",
        exp.DType.FLOAT: "Float32",
        exp.DType.INT: "Int32",
        exp.DType.MEDIUMINT: "Int32",
        exp.DType.INT128: "Int128",
        exp.DType.INT256: "Int256",
        exp.DType.LOWCARDINALITY: "LowCardinality",
        exp.DType.MAP: "Map",
        exp.DType.NESTED: "Nested",
        exp.DType.NOTHING: "Nothing",
        exp.DType.SMALLINT: "Int16",
        exp.DType.STRUCT: "Tuple",
        exp.DType.TINYINT: "Int8",
        exp.DType.UBIGINT: "UInt64",
        exp.DType.UINT: "UInt32",
        exp.DType.UINT128: "UInt128",
        exp.DType.UINT256: "UInt256",
        exp.DType.USMALLINT: "UInt16",
        exp.DType.UTINYINT: "UInt8",
        exp.DType.IPV4: "IPv4",
        exp.DType.IPV6: "IPv6",
        exp.DType.POINT: "Point",
        exp.DType.RING: "Ring",
        exp.DType.LINESTRING: "LineString",
        exp.DType.MULTILINESTRING: "MultiLineString",
        exp.DType.POLYGON: "Polygon",
        exp.DType.MULTIPOLYGON: "MultiPolygon",
        exp.DType.AGGREGATEFUNCTION: "AggregateFunction",
        exp.DType.SIMPLEAGGREGATEFUNCTION: "SimpleAggregateFunction",
        exp.DType.DYNAMIC: "Dynamic",
    }

    TRANSFORMS = {
        **{k: v for k, v in generator.Generator.TRANSFORMS.items() if k != exp.AutoRefreshProperty},
        exp.AnyValue: rename_func("any"),
        exp.ApproxDistinct: rename_func("uniq"),
        exp.ArrayDistinct: rename_func("arrayDistinct"),
        exp.ArrayConcat: rename_func("arrayConcat"),
        exp.ArrayContains: rename_func("has"),
        exp.ArrayFilter: lambda self, e: self.func("arrayFilter", e.expression, e.this),
        exp.Transform: lambda self, e: self.func("arrayMap", e.expression, e.this),
        exp.ArrayRemove: remove_from_array_using_filter,
        exp.ArrayReverse: rename_func("arrayReverse"),
        exp.ArraySlice: rename_func("arraySlice"),
        exp.ArraySum: rename_func("arraySum"),
        exp.ArrayMax: rename_func("arrayMax"),
        exp.ArrayMin: rename_func("arrayMin"),
        exp.ArgMax: arg_max_or_min_no_count("argMax"),
        exp.ArgMin: arg_max_or_min_no_count("argMin"),
        exp.Array: inline_array_sql,
        exp.CityHash64: rename_func("cityHash64"),
        exp.CastToStrType: rename_func("CAST"),
        exp.CurrentDatabase: rename_func("CURRENT_DATABASE"),
        exp.CurrentSchemas: rename_func("CURRENT_SCHEMAS"),
        exp.CountIf: rename_func("countIf"),
        exp.CosineDistance: rename_func("cosineDistance"),
        exp.CompressColumnConstraint: lambda self, e: (
            f"CODEC({self.expressions(e, key='this', flat=True)})"
        ),
        exp.ComputedColumnConstraint: lambda self, e: (
            f"{'MATERIALIZED' if e.args.get('persisted') else 'ALIAS'} {self.sql(e, 'this')}"
        ),
        exp.CurrentDate: lambda self, e: self.func("CURRENT_DATE"),
        exp.CurrentVersion: rename_func("VERSION"),
        exp.DateAdd: _datetime_delta_sql("DATE_ADD"),
        exp.DateDiff: _datetime_delta_sql("DATE_DIFF"),
        exp.DateStrToDate: rename_func("toDate"),
        exp.DateSub: _datetime_delta_sql("DATE_SUB"),
        exp.Explode: rename_func("arrayJoin"),
        exp.FarmFingerprint: rename_func("farmFingerprint64"),
        exp.Final: lambda self, e: f"{self.sql(e, 'this')} FINAL",
        exp.IsNan: rename_func("isNaN"),
        exp.JarowinklerSimilarity: jarowinkler_similarity("jaroWinklerSimilarity"),
        exp.JSONCast: _json_cast_sql,
        exp.JSONExtract: json_extract_segments("JSONExtractString", quoted_index=False),
        exp.JSONExtractScalar: json_extract_segments("JSONExtractString", quoted_index=False),
        exp.JSONPathKey: json_path_key_only_name,
        exp.JSONPathRoot: lambda *_: "",
        exp.Length: length_or_char_length_sql,
        exp.Map: _map_sql,
        exp.Median: rename_func("median"),
        exp.Nullif: rename_func("nullIf"),
        exp.PartitionedByProperty: lambda self, e: f"PARTITION BY {self.sql(e, 'this')}",
        exp.Pivot: no_pivot_sql,
        exp.Quantile: _quantile_sql,
        exp.RegexpLike: lambda self, e: self.func("match", e.this, e.expression),
        exp.Rand: rename_func("randCanonical"),
        exp.StartsWith: rename_func("startsWith"),
        exp.Struct: rename_func("tuple"),
        exp.Trunc: rename_func("trunc"),
        exp.EndsWith: rename_func("endsWith"),
        exp.EuclideanDistance: rename_func("L2Distance"),
        exp.StrPosition: lambda self, e: strposition_sql(
            self,
            e,
            func_name="POSITION",
            supports_position=True,
            use_ansi_position=False,
        ),
        exp.TimeToStr: lambda self, e: self.func(
            "formatDateTime",
            e.this.this if isinstance(e.this, exp.TsOrDsToTimestamp) else e.this,
            self.format_time(e),
            e.args.get("zone"),
        ),
        exp.TimeStrToTime: _timestrtotime_sql,
        exp.TimestampAdd: _datetime_delta_sql("TIMESTAMP_ADD"),
        exp.TimestampSub: _datetime_delta_sql("TIMESTAMP_SUB"),
        exp.Typeof: rename_func("toTypeName"),
        exp.VarMap: _map_sql,
        exp.Xor: lambda self, e: self.func("xor", e.this, e.expression),
        exp.MD5Digest: rename_func("MD5"),
        exp.MD5: lambda self, e: self.func("LOWER", self.func("HEX", self.func("MD5", e.this))),
        exp.SHA: rename_func("SHA1"),
        exp.SHA1Digest: rename_func("SHA1"),
        exp.SHA2: sha256_sql,
        exp.SHA2Digest: sha2_digest_sql,
        exp.Split: lambda self, e: self.func(
            "splitByString", e.args.get("expression"), e.this, e.args.get("limit")
        ),
        exp.RegexpSplit: lambda self, e: self.func(
            "splitByRegexp", e.args.get("expression"), e.this, e.args.get("limit")
        ),
        exp.UnixToTime: _unix_to_time_sql,
        exp.Trim: lambda self, e: trim_sql(self, e, default_trim_type="BOTH"),
        exp.Variance: rename_func("varSamp"),
        exp.SchemaCommentProperty: lambda self, e: self.naked_property(e),
        exp.Stddev: rename_func("stddevSamp"),
        exp.Chr: rename_func("CHAR"),
        exp.Lag: lambda self, e: self.func(
            "lagInFrame", e.this, e.args.get("offset"), e.args.get("default")
        ),
        exp.Lead: lambda self, e: self.func(
            "leadInFrame", e.this, e.args.get("offset"), e.args.get("default")
        ),
        exp.Levenshtein: unsupported_args("ins_cost", "del_cost", "sub_cost", "max_dist")(
            rename_func("editDistance")
        ),
        exp.ParseDatetime: lambda self, e: self.func(
            "parseDateTime", e.this, e.args.get("format"), e.args.get("zone")
        ),
    }

    PROPERTIES_LOCATION = {
        **generator.Generator.PROPERTIES_LOCATION,
        exp.AutoRefreshProperty: exp.Properties.Location.POST_NAME,
        exp.DefinerProperty: exp.Properties.Location.POST_SCHEMA,
        exp.OnCluster: exp.Properties.Location.POST_NAME,
        exp.PartitionedByProperty: exp.Properties.Location.POST_SCHEMA,
        exp.ToTableProperty: exp.Properties.Location.POST_NAME,
        exp.UuidProperty: exp.Properties.Location.POST_NAME,
        exp.VolatileProperty: exp.Properties.Location.UNSUPPORTED,
    }

    # There's no list in docs, but it can be found in Clickhouse code
    # see `ClickHouse/src/Parsers/ParserCreate*.cpp`
    ON_CLUSTER_TARGETS: t.ClassVar = {
        "SCHEMA",  # Transpiled CREATE SCHEMA may have OnCluster property set
        "DATABASE",
        "TABLE",
        "VIEW",
        "DICTIONARY",
        "INDEX",
        "FUNCTION",
        "NAMED COLLECTION",
    }

    # https://clickhouse.com/docs/en/sql-reference/data-types/nullable
    NON_NULLABLE_TYPES: t.ClassVar = {
        exp.DType.ARRAY,
        exp.DType.MAP,
        exp.DType.STRUCT,
        exp.DType.POINT,
        exp.DType.RING,
        exp.DType.LINESTRING,
        exp.DType.MULTILINESTRING,
        exp.DType.POLYGON,
        exp.DType.MULTIPOLYGON,
    }

    def groupconcat_sql(self, expression: exp.GroupConcat) -> str:
        this = expression.this
        separator = expression.args.get("separator")

        if isinstance(this, exp.Limit) and this.this:
            limit = this
            this = limit.this.pop()
            return self.sql(
                exp.ParameterizedAgg(
                    this="groupConcat",
                    params=[this],
                    expressions=[separator, limit.expression],
                )
            )

        if separator:
            return self.sql(
                exp.ParameterizedAgg(
                    this="groupConcat",
                    params=[this],
                    expressions=[separator],
                )
            )

        return self.func("groupConcat", this)

    def offset_sql(self, expression: exp.Offset) -> str:
        offset = super().offset_sql(expression)

        # OFFSET ... FETCH syntax requires a "ROW" or "ROWS" keyword
        # https://clickhouse.com/docs/sql-reference/statements/select/offset
        parent = expression.parent
        if isinstance(parent, exp.Select) and isinstance(parent.args.get("limit"), exp.Fetch):
            offset = f"{offset} ROWS"

        return offset

    def strtodate_sql(self, expression: exp.StrToDate) -> str:
        strtodate_sql = self.function_fallback_sql(expression)

        if not isinstance(expression.parent, exp.Cast):
            # StrToDate returns DATEs in other dialects (eg. postgres), so
            # this branch aims to improve the transpilation to clickhouse
            return self.cast_sql(exp.cast(expression, "DATE"))

        return strtodate_sql

    def cast_sql(self, expression: exp.Cast, safe_prefix: str | None = None) -> str:
        this = expression.this

        if isinstance(this, exp.StrToDate) and expression.to == exp.DType.DATETIME.into_expr():
            return self.sql(this)

        return super().cast_sql(expression, safe_prefix=safe_prefix)

    def trycast_sql(self, expression: exp.TryCast) -> str:
        dtype = expression.to
        if not dtype.is_type(*self.NON_NULLABLE_TYPES, check_nullable=True):
            # Casting x into Nullable(T) appears to behave similarly to TRY_CAST(x AS T)
            dtype.set("nullable", True)

        return super().cast_sql(expression)

    def _jsonpathsubscript_sql(self, expression: exp.JSONPathSubscript) -> str:
        this = self.json_path_part(expression.this)
        return str(int(this) + 1) if is_int(this) else this

    def likeproperty_sql(self, expression: exp.LikeProperty) -> str:
        return f"AS {self.sql(expression, 'this')}"

    def _any_to_has(
        self,
        expression: exp.EQ | exp.NEQ,
        default: t.Callable[[t.Any], str],
        prefix: str = "",
    ) -> str:
        if isinstance(expression.left, exp.Any):
            arr = expression.left
            this = expression.right
        elif isinstance(expression.right, exp.Any):
            arr = expression.right
            this = expression.left
        else:
            return default(expression)

        return prefix + self.func("has", arr.this.unnest(), this)

    def eq_sql(self, expression: exp.EQ) -> str:
        return self._any_to_has(expression, super().eq_sql)

    def neq_sql(self, expression: exp.NEQ) -> str:
        return self._any_to_has(expression, super().neq_sql, "NOT ")

    def regexpilike_sql(self, expression: exp.RegexpILike) -> str:
        # Manually add a flag to make the search case-insensitive
        regex = self.func("CONCAT", "'(?i)'", expression.expression)
        return self.func("match", expression.this, regex)

    def datatype_sql(self, expression: exp.DataType) -> str:
        # String is the standard ClickHouse type, every other variant is just an alias.
        # Additionally, any supplied length parameter will be ignored.
        #
        # https://clickhouse.com/docs/en/sql-reference/data-types/string
        if expression.this in self.STRING_TYPE_MAPPING:
            dtype = "String"
        else:
            dtype = super().datatype_sql(expression)

        # This section changes the type to `Nullable(...)` if the following conditions hold:
        # - It's marked as nullable - this ensures we won't wrap ClickHouse types with `Nullable`
        #   and change their semantics
        # - It's not the key type of a `Map`. This is because ClickHouse enforces the following
        #   constraint: "Type of Map key must be a type, that can be represented by integer or
        #   String or FixedString (possibly LowCardinality) or UUID or IPv6"
        # - It's not a composite type, e.g. `Nullable(Array(...))` is not a valid type
        parent = expression.parent
        nullable = expression.args.get("nullable")
        if nullable is True or (
            nullable is None
            and not (
                isinstance(parent, exp.DataType)
                and parent.is_type(exp.DType.MAP, check_nullable=True)
                and expression.index in (None, 0)
            )
            and not expression.is_type(*self.NON_NULLABLE_TYPES, check_nullable=True)
        ):
            dtype = f"Nullable({dtype})"

        return dtype

    def cte_sql(self, expression: exp.CTE) -> str:
        if expression.args.get("scalar"):
            this = self.sql(expression, "this")
            alias = self.sql(expression, "alias")
            return f"{this} AS {alias}"

        return super().cte_sql(expression)

    def after_limit_modifiers(self, expression: exp.Expr) -> list[str]:
        return super().after_limit_modifiers(expression) + [
            (
                self.seg("SETTINGS ") + self.expressions(expression, key="settings", flat=True)
                if expression.args.get("settings")
                else ""
            ),
            (
                self.seg("FORMAT ") + self.sql(expression, "format")
                if expression.args.get("format")
                else ""
            ),
        ]

    def placeholder_sql(self, expression: exp.Placeholder) -> str:
        return f"{{{expression.name}: {self.sql(expression, 'kind')}}}"

    def oncluster_sql(self, expression: exp.OnCluster) -> str:
        return f"ON CLUSTER {self.sql(expression, 'this')}"

    def _refresh_interval_sql(self, expression: exp.Expr) -> str:
        if isinstance(expression, exp.Add):
            return f"{self._refresh_interval_sql(expression.this)} {self._refresh_interval_sql(expression.expression)}"
        return self.sql(expression.assert_is(exp.Interval))

    def autorefreshproperty_sql(self, expression: exp.AutoRefreshProperty) -> str:
        cadence = self.sql(expression, "cadence")
        interval = expression.this
        schedule = (
            f" {cadence} {self._refresh_interval_sql(interval)}" if cadence and interval else ""
        )
        offset = expression.args.get("offset")
        offset = f" OFFSET {self._refresh_interval_sql(offset)}" if offset else ""
        randomize = expression.args.get("randomize")
        randomize = f" RANDOMIZE FOR {self._refresh_interval_sql(randomize)}" if randomize else ""
        dependencies = self.expressions(expression, flat=True)
        dependencies = f" DEPENDS ON {dependencies}" if dependencies else ""
        settings = self.sql(expression, "settings")
        settings = f" {settings}" if settings else ""
        append = " APPEND" if expression.args.get("append") else ""

        return f"REFRESH{schedule}{offset}{randomize}{dependencies}{settings}{append}"

    def createable_sql(self, expression: exp.Create, locations: defaultdict) -> str:
        if expression.kind in self.ON_CLUSTER_TARGETS and locations.get(
            exp.Properties.Location.POST_NAME
        ):
            this_name = self.sql(
                expression.this if isinstance(expression.this, exp.Schema) else expression,
                "this",
            )
            this_properties = " ".join(
                [self.sql(prop) for prop in locations[exp.Properties.Location.POST_NAME]]
            )
            this_schema = self.schema_columns_sql(expression.this)
            this_schema = f"{self.sep()}{this_schema}" if this_schema else ""

            return f"{this_name}{self.sep()}{this_properties}{this_schema}"

        return super().createable_sql(expression, locations)

    def create_sql(self, expression: exp.Create) -> str:
        # The comment property comes last in CTAS statements, i.e. after the query
        query = expression.expression
        if isinstance(query, exp.Query):
            comment_prop = expression.find(exp.SchemaCommentProperty)
            if comment_prop:
                comment_prop.pop()
                query.replace(exp.paren(query))
        else:
            comment_prop = None

        create_sql = super().create_sql(expression)

        comment_sql = self.sql(comment_prop)
        comment_sql = f" {comment_sql}" if comment_sql else ""

        return f"{create_sql}{comment_sql}"

    def prewhere_sql(self, expression: exp.PreWhere) -> str:
        this = self.indent(self.sql(expression, "this"))
        return f"{self.seg('PREWHERE')}{self.sep()}{this}"

    def indexcolumnconstraint_sql(self, expression: exp.IndexColumnConstraint) -> str:
        this = self.sql(expression, "this")
        this = f" {this}" if this else ""
        expr = self.sql(expression, "expression")
        expr = f" {expr}" if expr else ""
        index_type = self.sql(expression, "index_type")
        index_type = f" TYPE {index_type}" if index_type else ""
        granularity = self.sql(expression, "granularity")
        granularity = f" GRANULARITY {granularity}" if granularity else ""

        return f"INDEX{this}{expr}{index_type}{granularity}"

    def partition_sql(self, expression: exp.Partition) -> str:
        return f"PARTITION {self.expressions(expression, flat=True)}"

    def partitionid_sql(self, expression: exp.PartitionId) -> str:
        return f"ID {self.sql(expression.this)}"

    def replacepartition_sql(self, expression: exp.ReplacePartition) -> str:
        return f"REPLACE {self.sql(expression.expression)} FROM {self.sql(expression, 'source')}"

    def projectiondef_sql(self, expression: exp.ProjectionDef) -> str:
        return f"PROJECTION {self.sql(expression.this)} {self.wrap(expression.expression)}"

    def nestedjsonselect_sql(self, expression: exp.NestedJSONSelect) -> str:
        return f"{self.sql(expression, 'this')}.^{self.sql(expression, 'expression')}"

    def is_sql(self, expression: exp.Is) -> str:
        is_sql = super().is_sql(expression)

        if isinstance(expression.parent, exp.Not):
            # value IS NOT NULL -> NOT (value IS NULL)
            is_sql = self.wrap(is_sql)

        return is_sql

    def in_sql(self, expression: exp.In) -> str:
        in_sql = super().in_sql(expression)

        if isinstance(expression.parent, exp.Not) and expression.args.get("is_global"):
            in_sql = in_sql.replace("GLOBAL IN", "GLOBAL NOT IN", 1)

        return in_sql

    def not_sql(self, expression: exp.Not) -> str:
        if isinstance(expression.this, exp.In):
            if expression.this.args.get("is_global"):
                # let `GLOBAL IN` child interpose `NOT`
                return self.sql(expression, "this")

            expression.set("this", exp.paren(expression.this, copy=False))

        return super().not_sql(expression)

    def values_sql(self, expression: exp.Values, values_as_table: bool = True) -> str:
        # If the VALUES clause contains tuples of expressions, we need to treat it
        # as a table since Clickhouse will automatically alias it as such.
        alias = expression.args.get("alias")

        if alias and alias.args.get("columns") and expression.expressions:
            values = expression.expressions[0].expressions
            values_as_table = any(isinstance(value, exp.Tuple) for value in values)
        else:
            values_as_table = True

        return super().values_sql(expression, values_as_table=values_as_table)

    def timestamptrunc_sql(self, expression: exp.DateTrunc | exp.TimestampTrunc) -> str:
        unit = weekstart_unit_to_str(self, expression)
        # https://clickhouse.com/docs/whats-new/changelog/2023#improvement
        if self.dialect.version < (23, 12) and unit and unit.is_string:
            unit = exp.Literal.string(unit.name.lower())
        return self.func("dateTrunc", unit, expression.this, expression.args.get("zone"))

    def datetrunc_sql(self, expression: exp.DateTrunc) -> str:
        return self.timestamptrunc_sql(expression)
